from .annotation import Bitmask, BoundingBox, Detection, Polygon
from .manager import DataLoader, Evaluator, Filter, Metadata
from .metric import Metric, MetricType

__all__ = [
    "Bitmask",
    "BoundingBox",
    "Detection",
    "Polygon",
    "Metric",
    "MetricType",
    "DataLoader",
    "Evaluator",
    "Filter",
    "Metadata",
]



from dataclasses import asdict, dataclass

import numpy as np
from numpy.typing import NDArray
from tqdm import tqdm

from valor_lite.exceptions import EmptyEvaluatorError, EmptyFilterError
from valor_lite.semantic_segmentation.annotation import Segmentation
from valor_lite.semantic_segmentation.computation import (
    compute_intermediate_confusion_matrices,
    compute_label_metadata,
    compute_metrics,
    filter_cache,
)
from valor_lite.semantic_segmentation.metric import Metric, MetricType
from valor_lite.semantic_segmentation.utilities import (
    unpack_precision_recall_iou_into_metric_lists,
)

"""
Usage
-----

manager = DataLoader()
manager.add_data(
    groundtruths=groundtruths,
    predictions=predictions,
)
evaluator = manager.finalize()

metrics = evaluator.evaluate()

f1_metrics = metrics[MetricType.F1]
accuracy_metrics = metrics[MetricType.Accuracy]

filter_mask = evaluator.create_filter(datum_ids=["uid1", "uid2"])
filtered_metrics = evaluator.evaluate(filter_mask=filter_mask)
"""


@dataclass
class Metadata:
    number_of_labels: int = 0
    number_of_pixels: int = 0
    number_of_datums: int = 0
    number_of_ground_truths: int = 0
    number_of_predictions: int = 0

    @classmethod
    def create(
        cls,
        confusion_matrices: NDArray[np.int64],
    ):
        if confusion_matrices.size == 0:
            return cls()
        return cls(
            number_of_labels=confusion_matrices.shape[1] - 1,
            number_of_pixels=confusion_matrices.sum(),
            number_of_datums=confusion_matrices.shape[0],
            number_of_ground_truths=confusion_matrices[:, 1:, :].sum(),
            number_of_predictions=confusion_matrices[:, :, 1:].sum(),
        )

    def to_dict(self) -> dict[str, int | bool]:
        return asdict(self)


@dataclass
class Filter:
    datum_mask: NDArray[np.bool_]
    label_mask: NDArray[np.bool_]
    metadata: Metadata

    def __post_init__(self):
        # validate datum mask
        if not self.datum_mask.any():
            raise EmptyFilterError("filter removes all datums")

        # validate label mask
        if self.label_mask.all():
            raise EmptyFilterError("filter removes all labels")


class Evaluator:
    """
    Segmentation Evaluator
    """

    def __init__(self):
        """Initializes evaluator caches."""
        # external references
        self.datum_id_to_index: dict[str, int] = {}
        self.index_to_datum_id: list[str] = []
        self.label_to_index: dict[str, int] = {}
        self.index_to_label: list[str] = []

        # internal caches
        self._confusion_matrices = np.array([], dtype=np.int64)
        self._label_metadata = np.array([], dtype=np.int64)
        self._metadata = Metadata()

    @property
    def metadata(self) -> Metadata:
        return self._metadata

    @property
    def ignored_prediction_labels(self) -> list[str]:
        """
        Prediction labels that are not present in the ground truth set.
        """
        glabels = set(np.where(self._label_metadata[:, 0] > 0)[0])
        plabels = set(np.where(self._label_metadata[:, 1] > 0)[0])
        return [
            self.index_to_label[label_id] for label_id in (plabels - glabels)
        ]

    @property
    def missing_prediction_labels(self) -> list[str]:
        """
        Ground truth labels that are not present in the prediction set.
        """
        glabels = set(np.where(self._label_metadata[:, 0] > 0)[0])
        plabels = set(np.where(self._label_metadata[:, 1] > 0)[0])
        return [
            self.index_to_label[label_id] for label_id in (glabels - plabels)
        ]

    def create_filter(
        self,
        datums: list[str] | NDArray[np.int64] | None = None,
        labels: list[str] | NDArray[np.int64] | None = None,
    ) -> Filter:
        """
        Creates a filter for use with the evaluator.

        Parameters
        ----------
        datums : list[str] | NDArray[int64], optional
            An optional list of string ids or array of indices representing datums.
        labels : list[str] | NDArray[int64], optional
            An optional list of labels or array of indices.

        Returns
        -------
        Filter
            The filter object containing a mask and metadata.
        """
        datum_mask = np.ones(self._confusion_matrices.shape[0], dtype=np.bool_)
        label_mask = np.zeros(
            self.metadata.number_of_labels + 1, dtype=np.bool_
        )

        if datums is not None:
            # convert to indices
            if isinstance(datums, list):
                datums = np.array(
                    [self.datum_id_to_index[uid] for uid in datums],
                    dtype=np.int64,
                )

            # validate indices
            if datums.size == 0:
                raise EmptyFilterError(
                    "filter removes all datums"
                )  # validate indices
            elif datums.min() < 0:
                raise ValueError(
                    f"datum index cannot be negative '{datums.min()}'"
                )
            elif datums.max() >= len(self.index_to_datum_id):
                raise ValueError(
                    f"datum index cannot exceed total number of datums '{datums.max()}'"
                )

            # apply to mask
            datums.sort()
            mask_valid_datums = (
                np.arange(self._confusion_matrices.shape[0]).reshape(-1, 1)
                == datums.reshape(1, -1)
            ).any(axis=1)
            datum_mask[~mask_valid_datums] = False

        if labels is not None:
            # convert to indices
            if isinstance(labels, list):
                labels = np.array(
                    [self.label_to_index[label] for label in labels],
                    dtype=np.int64,
                )

            # validate indices
            if labels.size == 0:
                raise EmptyFilterError("filter removes all labels")
            elif labels.min() < 0:
                raise ValueError(
                    f"label index cannot be negative '{labels.min()}'"
                )
            elif labels.max() >= len(self.index_to_label):
                raise ValueError(
                    f"label index cannot exceed total number of labels '{labels.max()}'"
                )

            # apply to mask
            labels = np.concatenate([labels, np.array([-1])])
            label_range = np.arange(self.metadata.number_of_labels + 1) - 1
            mask_valid_labels = (
                label_range.reshape(-1, 1) == labels.reshape(1, -1)
            ).any(axis=1)
            label_mask[~mask_valid_labels] = True

        filtered_confusion_matrices, _ = filter_cache(
            confusion_matrices=self._confusion_matrices.copy(),
            datum_mask=datum_mask,
            label_mask=label_mask,
            number_of_labels=self.metadata.number_of_labels,
        )

        return Filter(
            datum_mask=datum_mask,
            label_mask=label_mask,
            metadata=Metadata.create(
                confusion_matrices=filtered_confusion_matrices,
            ),
        )

    def filter(
        self, filter_: Filter
    ) -> tuple[NDArray[np.int64], NDArray[np.int64]]:
        """
        Performs the filter operation over the internal cache.

        Parameters
        ----------
        filter_ : Filter
            An object describing the filter operation.

        Returns
        -------
        NDArray[int64]
            Filtered confusion matrices.
        NDArray[int64]
            Filtered label metadata
        """
        return filter_cache(
            confusion_matrices=self._confusion_matrices.copy(),
            datum_mask=filter_.datum_mask,
            label_mask=filter_.label_mask,
            number_of_labels=self.metadata.number_of_labels,
        )

    def compute_precision_recall_iou(
        self, filter_: Filter | None = None
    ) -> dict[MetricType, list]:
        """
        Performs an evaluation and returns metrics.

        Returns
        -------
        dict[MetricType, list]
            A dictionary mapping MetricType enumerations to lists of computed metrics.
        """
        if filter_ is not None:
            confusion_matrices, label_metadata = self.filter(filter_)
            n_pixels = filter_.metadata.number_of_pixels
        else:
            confusion_matrices = self._confusion_matrices
            label_metadata = self._label_metadata
            n_pixels = self.metadata.number_of_pixels

        results = compute_metrics(
            confusion_matrices=confusion_matrices,
            label_metadata=label_metadata,
            n_pixels=n_pixels,
        )
        return unpack_precision_recall_iou_into_metric_lists(
            results=results,
            label_metadata=label_metadata,
            index_to_label=self.index_to_label,
        )

    def evaluate(
        self, filter_: Filter | None = None
    ) -> dict[MetricType, list[Metric]]:
        """
        Computes all available metrics.

        Returns
        -------
        dict[MetricType, list[Metric]]
            Lists of metrics organized by metric type.
        """
        return self.compute_precision_recall_iou(filter_=filter_)


class DataLoader:
    """
    Segmentation DataLoader.
    """

    def __init__(self):
        self._evaluator = Evaluator()
        self.matrices = list()

    def _add_datum(self, uid: str) -> int:
        """
        Helper function for adding a datum to the cache.

        Parameters
        ----------
        uid : str
            The datum uid.

        Returns
        -------
        int
            The datum index.
        """
        if uid in self._evaluator.datum_id_to_index:
            raise ValueError(f"Datum with uid `{uid}` already exists.")
        index = len(self._evaluator.datum_id_to_index)
        self._evaluator.datum_id_to_index[uid] = index
        self._evaluator.index_to_datum_id.append(uid)
        return index

    def _add_label(self, label: str) -> int:
        """
        Helper function for adding a label to the cache.

        Parameters
        ----------
        label : str
            A string label.

        Returns
        -------
        int
            The label's index.
        """
        if label not in self._evaluator.label_to_index:
            label_id = len(self._evaluator.index_to_label)
            self._evaluator.label_to_index[label] = label_id
            self._evaluator.index_to_label.append(label)
        return self._evaluator.label_to_index[label]

    def add_data(
        self,
        segmentations: list[Segmentation],
        show_progress: bool = False,
    ):
        """
        Adds segmentations to the cache.

        Parameters
        ----------
        segmentations : list[Segmentation]
            A list of Segmentation objects.
        show_progress : bool, default=False
            Toggle for tqdm progress bar.
        """

        disable_tqdm = not show_progress
        for segmentation in tqdm(segmentations, disable=disable_tqdm):
            # update datum cache
            self._add_datum(segmentation.uid)

            groundtruth_labels = -1 * np.ones(
                len(segmentation.groundtruths), dtype=np.int64
            )
            for idx, groundtruth in enumerate(segmentation.groundtruths):
                label_idx = self._add_label(groundtruth.label)
                groundtruth_labels[idx] = label_idx

            prediction_labels = -1 * np.ones(
                len(segmentation.predictions), dtype=np.int64
            )
            for idx, prediction in enumerate(segmentation.predictions):
                label_idx = self._add_label(prediction.label)
                prediction_labels[idx] = label_idx

            if segmentation.groundtruths:
                combined_groundtruths = np.stack(
                    [
                        groundtruth.mask.flatten()
                        for groundtruth in segmentation.groundtruths
                    ],
                    axis=0,
                )
            else:
                combined_groundtruths = np.zeros(
                    (1, segmentation.shape[0] * segmentation.shape[1]),
                    dtype=np.bool_,
                )

            if segmentation.predictions:
                combined_predictions = np.stack(
                    [
                        prediction.mask.flatten()
                        for prediction in segmentation.predictions
                    ],
                    axis=0,
                )
            else:
                combined_predictions = np.zeros(
                    (1, segmentation.shape[0] * segmentation.shape[1]),
                    dtype=np.bool_,
                )

            self.matrices.append(
                compute_intermediate_confusion_matrices(
                    groundtruths=combined_groundtruths,
                    predictions=combined_predictions,
                    groundtruth_labels=groundtruth_labels,
                    prediction_labels=prediction_labels,
                    n_labels=len(self._evaluator.index_to_label),
                )
            )

    def finalize(self) -> Evaluator:
        """
        Performs data finalization and some preprocessing steps.

        Returns
        -------
        Evaluator
            A ready-to-use evaluator object.
        """

        if len(self.matrices) == 0:
            raise EmptyEvaluatorError()

        n_labels = len(self._evaluator.index_to_label)
        n_datums = len(self._evaluator.index_to_datum_id)
        self._evaluator._confusion_matrices = np.zeros(
            (n_datums, n_labels + 1, n_labels + 1), dtype=np.int64
        )
        for idx, matrix in enumerate(self.matrices):
            h, w = matrix.shape
            self._evaluator._confusion_matrices[idx, :h, :w] = matrix
        self._evaluator._label_metadata = compute_label_metadata(
            confusion_matrices=self._evaluator._confusion_matrices,
            n_labels=n_labels,
        )
        self._evaluator._metadata = Metadata.create(
            confusion_matrices=self._evaluator._confusion_matrices,
        )
        return self._evaluator



from collections import defaultdict

import numpy as np
from numpy.typing import NDArray

from valor_lite.object_detection.computation import PairClassification
from valor_lite.object_detection.metric import Metric, MetricType


def unpack_precision_recall_into_metric_lists(
    results: tuple[
        tuple[
            NDArray[np.float64],
            NDArray[np.float64],
        ],
        tuple[
            NDArray[np.float64],
            NDArray[np.float64],
        ],
        NDArray[np.float64],
        NDArray[np.float64],
    ],
    iou_thresholds: list[float],
    score_thresholds: list[float],
    index_to_label: list[str],
    label_metadata: NDArray[np.int32],
):
    (
        (
            average_precision,
            mean_average_precision,
        ),
        (
            average_recall,
            mean_average_recall,
        ),
        precision_recall,
        pr_curves,
    ) = results

    metrics = defaultdict(list)

    metrics[MetricType.AP] = [
        Metric.average_precision(
            value=float(average_precision[iou_idx][label_idx]),
            iou_threshold=iou_threshold,
            label=label,
        )
        for iou_idx, iou_threshold in enumerate(iou_thresholds)
        for label_idx, label in enumerate(index_to_label)
        if int(label_metadata[label_idx, 0]) > 0
    ]

    metrics[MetricType.mAP] = [
        Metric.mean_average_precision(
            value=float(mean_average_precision[iou_idx]),
            iou_threshold=iou_threshold,
        )
        for iou_idx, iou_threshold in enumerate(iou_thresholds)
    ]

    # TODO - (c.zaloom) will be removed in the future
    metrics[MetricType.APAveragedOverIOUs] = [
        Metric.average_precision_averaged_over_IOUs(
            value=float(average_precision.mean(axis=0)[label_idx]),
            iou_thresholds=iou_thresholds,
            label=label,
        )
        for label_idx, label in enumerate(index_to_label)
        if int(label_metadata[label_idx, 0]) > 0
    ]

    # TODO - (c.zaloom) will be removed in the future
    metrics[MetricType.mAPAveragedOverIOUs] = [
        Metric.mean_average_precision_averaged_over_IOUs(
            value=float(mean_average_precision.mean()),
            iou_thresholds=iou_thresholds,
        )
    ]

    metrics[MetricType.AR] = [
        Metric.average_recall(
            value=float(average_recall[score_idx, label_idx]),
            iou_thresholds=iou_thresholds,
            score_threshold=score_threshold,
            label=label,
        )
        for score_idx, score_threshold in enumerate(score_thresholds)
        for label_idx, label in enumerate(index_to_label)
        if int(label_metadata[label_idx, 0]) > 0
    ]

    metrics[MetricType.mAR] = [
        Metric.mean_average_recall(
            value=float(mean_average_recall[score_idx]),
            iou_thresholds=iou_thresholds,
            score_threshold=score_threshold,
        )
        for score_idx, score_threshold in enumerate(score_thresholds)
    ]

    # TODO - (c.zaloom) will be removed in the future
    metrics[MetricType.ARAveragedOverScores] = [
        Metric.average_recall_averaged_over_scores(
            value=float(average_recall.mean(axis=0)[label_idx]),
            score_thresholds=score_thresholds,
            iou_thresholds=iou_thresholds,
            label=label,
        )
        for label_idx, label in enumerate(index_to_label)
        if int(label_metadata[label_idx, 0]) > 0
    ]

    # TODO - (c.zaloom) will be removed in the future
    metrics[MetricType.mARAveragedOverScores] = [
        Metric.mean_average_recall_averaged_over_scores(
            value=float(mean_average_recall.mean()),
            score_thresholds=score_thresholds,
            iou_thresholds=iou_thresholds,
        )
    ]

    metrics[MetricType.PrecisionRecallCurve] = [
        Metric.precision_recall_curve(
            precisions=pr_curves[iou_idx, label_idx, :, 0].tolist(),  # type: ignore[reportArgumentType]
            scores=pr_curves[iou_idx, label_idx, :, 1].tolist(),  # type: ignore[reportArgumentType]
            iou_threshold=iou_threshold,
            label=label,
        )
        for iou_idx, iou_threshold in enumerate(iou_thresholds)
        for label_idx, label in enumerate(index_to_label)
        if label_metadata[label_idx, 0] > 0
    ]

    for label_idx, label in enumerate(index_to_label):
        if label_metadata[label_idx, 0] == 0:
            continue

        for score_idx, score_threshold in enumerate(score_thresholds):
            for iou_idx, iou_threshold in enumerate(iou_thresholds):

                row = precision_recall[iou_idx, score_idx, label_idx, :]
                kwargs = {
                    "label": label,
                    "iou_threshold": iou_threshold,
                    "score_threshold": score_threshold,
                }
                metrics[MetricType.Counts].append(
                    Metric.counts(
                        tp=int(row[0]),
                        fp=int(row[1]),
                        fn=int(row[2]),
                        **kwargs,
                    )
                )

                metrics[MetricType.Precision].append(
                    Metric.precision(
                        value=float(row[3]),
                        **kwargs,
                    )
                )
                metrics[MetricType.Recall].append(
                    Metric.recall(
                        value=float(row[4]),
                        **kwargs,
                    )
                )
                metrics[MetricType.F1].append(
                    Metric.f1_score(
                        value=float(row[5]),
                        **kwargs,
                    )
                )

    return metrics


def _create_empty_confusion_matrix(index_to_labels: list[str]):
    unmatched_ground_truths = dict()
    unmatched_predictions = dict()
    confusion_matrix = dict()
    for label in index_to_labels:
        unmatched_ground_truths[label] = {"count": 0, "examples": []}
        unmatched_predictions[label] = {"count": 0, "examples": []}
        confusion_matrix[label] = {}
        for plabel in index_to_labels:
            confusion_matrix[label][plabel] = {"count": 0, "examples": []}
    return (
        confusion_matrix,
        unmatched_predictions,
        unmatched_ground_truths,
    )


def _unpack_confusion_matrix(
    ids: NDArray[np.int32],
    mask_matched: NDArray[np.bool_],
    mask_fp_unmatched: NDArray[np.bool_],
    mask_fn_unmatched: NDArray[np.bool_],
    index_to_datum_id: list[str],
    index_to_groundtruth_id: list[str],
    index_to_prediction_id: list[str],
    index_to_label: list[str],
    iou_threhsold: float,
    score_threshold: float,
):
    (
        confusion_matrix,
        unmatched_predictions,
        unmatched_ground_truths,
    ) = _create_empty_confusion_matrix(index_to_label)

    unique_matches = np.unique(
        ids[np.ix_(mask_matched, (0, 1, 2, 3, 4))], axis=0  # type: ignore - numpy ix_ typing
    )
    unique_unmatched_predictions = np.unique(
        ids[np.ix_(mask_fp_unmatched, (0, 2, 4))], axis=0  # type: ignore - numpy ix_ typing
    )
    unique_unmatched_groundtruths = np.unique(
        ids[np.ix_(mask_fn_unmatched, (0, 1, 3))], axis=0  # type: ignore - numpy ix_ typing
    )

    n_matched = unique_matches.shape[0]
    n_unmatched_predictions = unique_unmatched_predictions.shape[0]
    n_unmatched_groundtruths = unique_unmatched_groundtruths.shape[0]
    n_max = max(n_matched, n_unmatched_groundtruths, n_unmatched_predictions)

    for idx in range(n_max):
        if idx < n_unmatched_groundtruths:
            label = index_to_label[unique_unmatched_groundtruths[idx, 2]]
            unmatched_ground_truths[label]["count"] += 1
            unmatched_ground_truths[label]["examples"].append(
                {
                    "datum_id": index_to_datum_id[
                        unique_unmatched_groundtruths[idx, 0]
                    ],
                    "ground_truth_id": index_to_groundtruth_id[
                        unique_unmatched_groundtruths[idx, 1]
                    ],
                }
            )
        if idx < n_unmatched_predictions:
            label = index_to_label[unique_unmatched_predictions[idx, 2]]
            unmatched_predictions[label]["count"] += 1
            unmatched_predictions[label]["examples"].append(
                {
                    "datum_id": index_to_datum_id[
                        unique_unmatched_predictions[idx, 0]
                    ],
                    "prediction_id": index_to_prediction_id[
                        unique_unmatched_predictions[idx, 1]
                    ],
                }
            )
        if idx < n_matched:
            glabel = index_to_label[unique_matches[idx, 3]]
            plabel = index_to_label[unique_matches[idx, 4]]
            confusion_matrix[glabel][plabel]["count"] += 1
            confusion_matrix[glabel][plabel]["examples"].append(
                {
                    "datum_id": index_to_datum_id[unique_matches[idx, 0]],
                    "ground_truth_id": index_to_groundtruth_id[
                        unique_matches[idx, 1]
                    ],
                    "prediction_id": index_to_prediction_id[
                        unique_matches[idx, 2]
                    ],
                }
            )

    return Metric.confusion_matrix(
        confusion_matrix=confusion_matrix,
        unmatched_ground_truths=unmatched_ground_truths,
        unmatched_predictions=unmatched_predictions,
        iou_threshold=iou_threhsold,
        score_threshold=score_threshold,
    )


def unpack_confusion_matrix_into_metric_list(
    results: NDArray[np.uint8],
    detailed_pairs: NDArray[np.float64],
    iou_thresholds: list[float],
    score_thresholds: list[float],
    index_to_datum_id: list[str],
    index_to_groundtruth_id: list[str],
    index_to_prediction_id: list[str],
    index_to_label: list[str],
) -> list[Metric]:

    ids = detailed_pairs[:, :5].astype(np.int32)

    mask_matched = (
        np.bitwise_and(
            results, PairClassification.TP | PairClassification.FP_FN_MISCLF
        )
        > 0
    )
    mask_fp_unmatched = (
        np.bitwise_and(results, PairClassification.FP_UNMATCHED) > 0
    )
    mask_fn_unmatched = (
        np.bitwise_and(results, PairClassification.FN_UNMATCHED) > 0
    )

    return [
        _unpack_confusion_matrix(
            ids=ids,
            mask_matched=mask_matched[iou_idx, score_idx],
            mask_fp_unmatched=mask_fp_unmatched[iou_idx, score_idx],
            mask_fn_unmatched=mask_fn_unmatched[iou_idx, score_idx],
            index_to_datum_id=index_to_datum_id,
            index_to_groundtruth_id=index_to_groundtruth_id,
            index_to_prediction_id=index_to_prediction_id,
            index_to_label=index_to_label,
            iou_threhsold=iou_threshold,
            score_threshold=score_threshold,
        )
        for iou_idx, iou_threshold in enumerate(iou_thresholds)
        for score_idx, score_threshold in enumerate(score_thresholds)
        if (results[iou_idx, score_idx] != -1).any()
    ]



import numpy as np

from valor_lite.profiling import create_runtime_profiler
from valor_lite.semantic_segmentation import Bitmask, DataLoader, Segmentation


def generate_segmentation(
    datum_uid: str,
    number_of_unique_labels: int,
    mask_height: int,
    mask_width: int,
) -> Segmentation:
    """
    Generates a semantic segmentation annotation.

    Parameters
    ----------
    datum_uid : str
        The datum UID for the generated segmentation.
    number_of_unique_labels : int
        The number of unique labels.
    mask_height : int
        The height of the mask in pixels.
    mask_width : int
        The width of the mask in pixels.

    Returns
    -------
    Segmentation
        A generated semantic segmenatation annotation.
    """

    if number_of_unique_labels > 1:
        common_proba = 0.4 / (number_of_unique_labels - 1)
        min_proba = min(common_proba, 0.1)
        labels = [str(i) for i in range(number_of_unique_labels)] + [None]
        proba = (
            [0.5]
            + [common_proba for _ in range(number_of_unique_labels - 1)]
            + [0.1]
        )
    elif number_of_unique_labels == 1:
        labels = ["0", None]
        proba = [0.9, 0.1]
        min_proba = 0.1
    else:
        raise ValueError(
            "The number of unique labels should be greater than zero."
        )

    probabilities = np.array(proba, dtype=np.float64)
    weights = (probabilities / min_proba).astype(np.int32)

    indices = np.random.choice(
        np.arange(len(weights)),
        size=(mask_height * 2, mask_width),
        p=probabilities,
    )

    N = len(labels)

    masks = np.arange(N)[:, None, None] == indices

    gts = []
    pds = []
    for lidx in range(N):
        label = labels[lidx]
        if label is None:
            continue
        gts.append(
            Bitmask(
                mask=masks[lidx, :mask_height, :],
                label=label,
            )
        )
        pds.append(
            Bitmask(
                mask=masks[lidx, mask_height:, :],
                label=label,
            )
        )

    return Segmentation(
        uid=datum_uid,
        groundtruths=gts,
        predictions=pds,
        shape=(mask_height, mask_width),
    )


def benchmark_add_data(
    n_labels: int,
    shape: tuple[int, int],
    time_limit: float | None,
    repeat: int = 1,
) -> float:
    """
    Benchmarks 'Dataloader.add_data' for semantic segmentation.

    Parameters
    ----------
    n_labels : int
        The number of unique labels to generate.
    shape : tuple[int, int]
        The size (h,w) of the mask to generate.
    time_limit : float, optional
        An optional time limit to constrain the benchmark.
    repeat : int
        The number of times to run the benchmark to produce a runtime average.

    Returns
    -------
    float
        The average runtime.
    """

    profile = create_runtime_profiler(
        time_limit=time_limit,
        repeat=repeat,
    )

    elapsed = 0
    for _ in range(repeat):
        data = generate_segmentation(
            datum_uid="uid",
            number_of_unique_labels=n_labels,
            mask_height=shape[0],
            mask_width=shape[1],
        )
        loader = DataLoader()
        elapsed += profile(loader.add_data)([data])
    return elapsed / repeat


def benchmark_finalize(
    n_datums: int,
    n_labels: int,
    time_limit: float | None,
    repeat: int = 1,
):
    """
    Benchmarks 'Dataloader.finalize' for semantic segmentation.

    Parameters
    ----------
    n_datums : int
        The number of datums to generate.
    n_labels : int
        The number of unique labels to generate.
    time_limit : float, optional
        An optional time limit to constrain the benchmark.
    repeat : int
        The number of times to run the benchmark to produce a runtime average.

    Returns
    -------
    float
        The average runtime.
    """

    profile = create_runtime_profiler(
        time_limit=time_limit,
        repeat=repeat,
    )

    elapsed = 0
    for _ in range(repeat):

        data = [
            generate_segmentation(
                datum_uid=str(i),
                number_of_unique_labels=n_labels,
                mask_height=5,
                mask_width=5,
            )
            for i in range(10)
        ]
        loader = DataLoader()
        for datum_idx in range(n_datums):
            segmentation = data[datum_idx % 10]
            segmentation.uid = str(datum_idx)
            loader.add_data([segmentation])
        elapsed += profile(loader.finalize)()
    return elapsed / repeat


def benchmark_evaluate(
    n_datums: int,
    n_labels: int,
    time_limit: float | None,
    repeat: int = 1,
):
    """
    Benchmarks 'Evaluator.evaluate' for semantic segmentation.

    Parameters
    ----------
    n_datums : int
        The number of datums to generate.
    n_labels : int
        The number of unique labels to generate.
    time_limit : float, optional
        An optional time limit to constrain the benchmark.
    repeat : int
        The number of times to run the benchmark to produce a runtime average.

    Returns
    -------
    float
        The average runtime.
    """

    profile = create_runtime_profiler(
        time_limit=time_limit,
        repeat=repeat,
    )

    elapsed = 0
    for _ in range(repeat):

        data = [
            generate_segmentation(
                datum_uid=str(i),
                number_of_unique_labels=n_labels,
                mask_height=5,
                mask_width=5,
            )
            for i in range(10)
        ]
        loader = DataLoader()
        for datum_idx in range(n_datums):
            segmentation = data[datum_idx % 10]
            segmentation.uid = str(datum_idx)
            loader.add_data([segmentation])
        evaluator = loader.finalize()
        elapsed += profile(evaluator.evaluate)()
    return elapsed / repeat



from dataclasses import dataclass, field

import numpy as np
from numpy.typing import NDArray


@dataclass
class Bitmask:
    """
    Represents a binary mask with an associated semantic label.

    Parameters
    ----------
    mask : NDArray[np.bool_]
        A NumPy array of boolean values representing the mask.
    label : str
        The semantic label associated with the mask.

    Examples
    --------
    >>> import numpy as np
    >>> mask = np.array([[True, False], [False, True]], dtype=np.bool_)
    >>> bitmask = Bitmask(mask=mask, label='ocean')
    """

    mask: NDArray[np.bool_]
    label: str

    def __post_init__(self):
        if self.mask.dtype != np.bool_:
            raise ValueError(
                f"Bitmask recieved mask with dtype '{self.mask.dtype}'."
            )


@dataclass
class Segmentation:
    """
    Segmentation data structure holding ground truth and prediction bitmasks for semantic segmentation tasks.

    Parameters
    ----------
    uid : str
        Unique identifier for the image or sample.
    groundtruths : List[Bitmask]
        List of ground truth bitmasks.
    predictions : List[Bitmask]
        List of predicted bitmasks.
    shape : tuple of int, optional
        The shape of the segmentation masks. This is set automatically after initialization.
    size : int, optional
        The total number of pixels in the masks. This is set automatically after initialization.

    Examples
    --------
    >>> import numpy as np
    >>> mask1 = np.array([[True, False], [False, True]], dtype=np.bool_)
    >>> groundtruth = Bitmask(mask=mask1, label='object')
    >>> mask2 = np.array([[False, True], [True, False]], dtype=np.bool_)
    >>> prediction = Bitmask(mask=mask2, label='object')
    >>> segmentation = Segmentation(
    ...     uid='123',
    ...     groundtruths=[groundtruth],
    ...     predictions=[prediction]
    ... )
    """

    uid: str
    groundtruths: list[Bitmask]
    predictions: list[Bitmask]
    shape: tuple[int, ...]
    size: int = field(default=0)

    def __post_init__(self):

        if len(self.shape) != 2 or self.shape[0] <= 0 or self.shape[1] <= 0:
            raise ValueError(
                f"segmentations must be 2-dimensional and have non-zero dimensions. Recieved shape '{self.shape}'"
            )
        self.size = self.shape[0] * self.shape[1]

        mask_accumulation = None
        for groundtruth in self.groundtruths:
            if self.shape != groundtruth.mask.shape:
                raise ValueError(
                    f"ground truth masks for datum '{self.uid}' should have shape '{self.shape}'. Received mask with shape '{groundtruth.mask.shape}'"
                )

            if mask_accumulation is None:
                mask_accumulation = groundtruth.mask.copy()
            elif np.logical_and(mask_accumulation, groundtruth.mask).any():
                raise ValueError("ground truth masks cannot overlap")
            else:
                mask_accumulation = mask_accumulation | groundtruth.mask

        mask_accumulation = None
        for prediction in self.predictions:
            if self.shape != prediction.mask.shape:
                raise ValueError(
                    f"prediction masks for datum '{self.uid}' should have shape '{self.shape}'. Received mask with shape '{prediction.mask.shape}'"
                )

            if mask_accumulation is None:
                mask_accumulation = prediction.mask.copy()
            elif np.logical_and(mask_accumulation, prediction.mask).any():
                raise ValueError("prediction masks cannot overlap")
            else:
                mask_accumulation = mask_accumulation | prediction.mask



from collections import defaultdict

import numpy as np
from numpy.typing import NDArray

from valor_lite.classification.computation import PairClassification
from valor_lite.classification.metric import Metric, MetricType


def unpack_precision_recall_rocauc_into_metric_lists(
    results: tuple[
        NDArray[np.int32],
        NDArray[np.float64],
        NDArray[np.float64],
        NDArray[np.float64],
        NDArray[np.float64],
        NDArray[np.float64],
        float,
    ],
    score_thresholds: list[float],
    hardmax: bool,
    label_metadata: NDArray[np.int32],
    index_to_label: list[str],
) -> dict[MetricType, list[Metric]]:
    (
        counts,
        precision,
        recall,
        accuracy,
        f1_score,
        rocauc,
        mean_rocauc,
    ) = results

    metrics = defaultdict(list)

    metrics[MetricType.ROCAUC] = [
        Metric.roc_auc(
            value=float(rocauc[label_idx]),
            label=label,
        )
        for label_idx, label in enumerate(index_to_label)
        if label_metadata[label_idx, 0] > 0
    ]

    metrics[MetricType.mROCAUC] = [
        Metric.mean_roc_auc(
            value=float(mean_rocauc),
        )
    ]

    metrics[MetricType.Accuracy] = [
        Metric.accuracy(
            value=float(accuracy[score_idx]),
            score_threshold=score_threshold,
            hardmax=hardmax,
        )
        for score_idx, score_threshold in enumerate(score_thresholds)
    ]

    for label_idx, label in enumerate(index_to_label):
        for score_idx, score_threshold in enumerate(score_thresholds):

            kwargs = {
                "label": label,
                "hardmax": hardmax,
                "score_threshold": score_threshold,
            }
            row = counts[:, label_idx]
            metrics[MetricType.Counts].append(
                Metric.counts(
                    tp=int(row[score_idx, 0]),
                    fp=int(row[score_idx, 1]),
                    fn=int(row[score_idx, 2]),
                    tn=int(row[score_idx, 3]),
                    **kwargs,
                )
            )

            # if no groundtruths exists for a label, skip it.
            if label_metadata[label_idx, 0] == 0:
                continue

            metrics[MetricType.Precision].append(
                Metric.precision(
                    value=float(precision[score_idx, label_idx]),
                    **kwargs,
                )
            )
            metrics[MetricType.Recall].append(
                Metric.recall(
                    value=float(recall[score_idx, label_idx]),
                    **kwargs,
                )
            )
            metrics[MetricType.F1].append(
                Metric.f1_score(
                    value=float(f1_score[score_idx, label_idx]),
                    **kwargs,
                )
            )
    return metrics


def _create_empty_confusion_matrix(index_to_labels: list[str]):
    unmatched_ground_truths = dict()
    confusion_matrix = dict()
    for label in index_to_labels:
        unmatched_ground_truths[label] = {"count": 0, "examples": []}
        confusion_matrix[label] = {}
        for plabel in index_to_labels:
            confusion_matrix[label][plabel] = {"count": 0, "examples": []}
    return (
        confusion_matrix,
        unmatched_ground_truths,
    )


def _unpack_confusion_matrix(
    ids: NDArray[np.int32],
    scores: NDArray[np.float64],
    mask_matched: NDArray[np.bool_],
    mask_fn_unmatched: NDArray[np.bool_],
    index_to_datum_id: list[str],
    index_to_label: list[str],
    score_threshold: float,
):
    (
        confusion_matrix,
        unmatched_ground_truths,
    ) = _create_empty_confusion_matrix(index_to_label)

    unique_matches, unique_match_indices = np.unique(
        ids[np.ix_(mask_matched, (0, 1, 2))],  # type: ignore - numpy ix_ typing
        axis=0,
        return_index=True,
    )
    (
        unique_unmatched_groundtruths,
        unique_unmatched_groundtruth_indices,
    ) = np.unique(
        ids[np.ix_(mask_fn_unmatched, (0, 1))],  # type: ignore - numpy ix_ typing
        axis=0,
        return_index=True,
    )

    n_matched = unique_matches.shape[0]
    n_unmatched_groundtruths = unique_unmatched_groundtruths.shape[0]
    n_max = max(n_matched, n_unmatched_groundtruths)

    for idx in range(n_max):
        if idx < n_matched:
            glabel = index_to_label[unique_matches[idx, 1]]
            plabel = index_to_label[unique_matches[idx, 2]]
            confusion_matrix[glabel][plabel]["count"] += 1
            confusion_matrix[glabel][plabel]["examples"].append(
                {
                    "datum_id": index_to_datum_id[unique_matches[idx, 0]],
                    "score": float(scores[unique_match_indices[idx]]),
                }
            )
        if idx < n_unmatched_groundtruths:
            label = index_to_label[unique_unmatched_groundtruths[idx, 1]]
            unmatched_ground_truths[label]["count"] += 1
            unmatched_ground_truths[label]["examples"].append(
                {
                    "datum_id": index_to_datum_id[
                        unique_unmatched_groundtruths[idx, 0]
                    ],
                }
            )

    return Metric.confusion_matrix(
        confusion_matrix=confusion_matrix,
        unmatched_ground_truths=unmatched_ground_truths,
        score_threshold=score_threshold,
    )


def unpack_confusion_matrix_into_metric_list(
    result: NDArray[np.uint8],
    detailed_pairs: NDArray[np.float64],
    score_thresholds: list[float],
    index_to_datum_id: list[str],
    index_to_label: list[str],
) -> list[Metric]:

    ids = detailed_pairs[:, :3].astype(np.int32)

    mask_matched = (
        np.bitwise_and(
            result, PairClassification.TP | PairClassification.FP_FN_MISCLF
        )
        > 0
    )
    mask_fn_unmatched = (
        np.bitwise_and(result, PairClassification.FN_UNMATCHED) > 0
    )

    return [
        _unpack_confusion_matrix(
            ids=ids,
            scores=detailed_pairs[:, 3],
            mask_matched=mask_matched[score_idx, :],
            mask_fn_unmatched=mask_fn_unmatched[score_idx, :],
            index_to_datum_id=index_to_datum_id,
            index_to_label=index_to_label,
            score_threshold=score_threshold,
        )
        for score_idx, score_threshold in enumerate(score_thresholds)
    ]



from .annotation import Bitmask, Segmentation
from .manager import DataLoader, Evaluator, Filter, Metadata
from .metric import Metric, MetricType

__all__ = [
    "DataLoader",
    "Evaluator",
    "Segmentation",
    "Bitmask",
    "Metric",
    "MetricType",
    "Filter",
    "Metadata",
]



import warnings
from dataclasses import asdict, dataclass

import numpy as np
from numpy.typing import NDArray
from tqdm import tqdm

from valor_lite.exceptions import (
    EmptyEvaluatorError,
    EmptyFilterError,
    InternalCacheError,
)
from valor_lite.object_detection.annotation import (
    Bitmask,
    BoundingBox,
    Detection,
    Polygon,
)
from valor_lite.object_detection.computation import (
    compute_bbox_iou,
    compute_bitmask_iou,
    compute_confusion_matrix,
    compute_label_metadata,
    compute_polygon_iou,
    compute_precion_recall,
    filter_cache,
    rank_pairs,
)
from valor_lite.object_detection.metric import Metric, MetricType
from valor_lite.object_detection.utilities import (
    unpack_confusion_matrix_into_metric_list,
    unpack_precision_recall_into_metric_lists,
)

"""
Usage
-----

loader = DataLoader()
loader.add_bounding_boxes(
    groundtruths=groundtruths,
    predictions=predictions,
)
evaluator = loader.finalize()

metrics = evaluator.evaluate(iou_thresholds=[0.5])

ap_metrics = metrics[MetricType.AP]
ar_metrics = metrics[MetricType.AR]

filter_mask = evaluator.create_filter(datum_uids=["uid1", "uid2"])
filtered_metrics = evaluator.evaluate(iou_thresholds=[0.5], filter_mask=filter_mask)
"""


@dataclass
class Metadata:
    number_of_datums: int = 0
    number_of_ground_truths: int = 0
    number_of_predictions: int = 0
    number_of_labels: int = 0

    @classmethod
    def create(
        cls,
        detailed_pairs: NDArray[np.float64],
        number_of_datums: int,
        number_of_labels: int,
    ):
        # count number of ground truths
        mask_valid_gts = detailed_pairs[:, 1] >= 0
        unique_ids = np.unique(
            detailed_pairs[np.ix_(mask_valid_gts, (0, 1))], axis=0  # type: ignore - np.ix_ typing
        )
        number_of_ground_truths = int(unique_ids.shape[0])

        # count number of predictions
        mask_valid_pds = detailed_pairs[:, 2] >= 0
        unique_ids = np.unique(
            detailed_pairs[np.ix_(mask_valid_pds, (0, 2))], axis=0  # type: ignore - np.ix_ typing
        )
        number_of_predictions = int(unique_ids.shape[0])

        return cls(
            number_of_datums=number_of_datums,
            number_of_ground_truths=number_of_ground_truths,
            number_of_predictions=number_of_predictions,
            number_of_labels=number_of_labels,
        )

    def to_dict(self) -> dict[str, int | bool]:
        return asdict(self)


@dataclass
class Filter:
    mask_datums: NDArray[np.bool_]
    mask_groundtruths: NDArray[np.bool_]
    mask_predictions: NDArray[np.bool_]
    metadata: Metadata

    def __post_init__(self):
        # validate datums mask
        if not self.mask_datums.any():
            raise EmptyFilterError("filter removes all datums")

        # validate annotation masks
        no_gts = self.mask_groundtruths.all()
        no_pds = self.mask_predictions.all()
        if no_gts and no_pds:
            raise EmptyFilterError("filter removes all annotations")
        elif no_gts:
            warnings.warn("filter removes all ground truths")
        elif no_pds:
            warnings.warn("filter removes all predictions")


class Evaluator:
    """
    Object Detection Evaluator
    """

    def __init__(self):

        # external reference
        self.datum_id_to_index: dict[str, int] = {}
        self.groundtruth_id_to_index: dict[str, int] = {}
        self.prediction_id_to_index: dict[str, int] = {}
        self.label_to_index: dict[str, int] = {}

        self.index_to_datum_id: list[str] = []
        self.index_to_groundtruth_id: list[str] = []
        self.index_to_prediction_id: list[str] = []
        self.index_to_label: list[str] = []

        # temporary cache
        self._temp_cache: list[NDArray[np.float64]] | None = []

        # internal cache
        self._detailed_pairs = np.array([[]], dtype=np.float64)
        self._ranked_pairs = np.array([[]], dtype=np.float64)
        self._label_metadata: NDArray[np.int32] = np.array([[]])
        self._metadata = Metadata()

    @property
    def ignored_prediction_labels(self) -> list[str]:
        """
        Prediction labels that are not present in the ground truth set.
        """
        glabels = set(np.where(self._label_metadata[:, 0] > 0)[0])
        plabels = set(np.where(self._label_metadata[:, 1] > 0)[0])
        return [
            self.index_to_label[label_id] for label_id in (plabels - glabels)
        ]

    @property
    def missing_prediction_labels(self) -> list[str]:
        """
        Ground truth labels that are not present in the prediction set.
        """
        glabels = set(np.where(self._label_metadata[:, 0] > 0)[0])
        plabels = set(np.where(self._label_metadata[:, 1] > 0)[0])
        return [
            self.index_to_label[label_id] for label_id in (glabels - plabels)
        ]

    @property
    def metadata(self) -> Metadata:
        """
        Evaluation metadata.
        """
        return self._metadata

    def create_filter(
        self,
        datums: list[str] | NDArray[np.int32] | None = None,
        groundtruths: list[str] | NDArray[np.int32] | None = None,
        predictions: list[str] | NDArray[np.int32] | None = None,
        labels: list[str] | NDArray[np.int32] | None = None,
    ) -> Filter:
        """
        Creates a filter object.

        Parameters
        ----------
        datum : list[str] | NDArray[int32], optional
            An optional list of string ids or indices representing datums to keep.
        groundtruth : list[str] | NDArray[int32], optional
            An optional list of string ids or indices representing ground truth annotations to keep.
        prediction : list[str] | NDArray[int32], optional
            An optional list of string ids or indices representing prediction annotations to keep.
        labels : list[str] | NDArray[int32], optional
            An optional list of labels or indices to keep.
        """
        mask_datums = np.ones(self._detailed_pairs.shape[0], dtype=np.bool_)

        # filter datums
        if datums is not None:
            # convert to indices
            if isinstance(datums, list):
                datums = np.array(
                    [self.datum_id_to_index[uid] for uid in datums],
                    dtype=np.int32,
                )

            # validate indices
            if datums.size == 0:
                raise EmptyFilterError(
                    "filter removes all datums"
                )  # validate indices
            elif datums.min() < 0:
                raise ValueError(
                    f"datum index cannot be negative '{datums.min()}'"
                )
            elif datums.max() >= len(self.index_to_datum_id):
                raise ValueError(
                    f"datum index cannot exceed total number of datums '{datums.max()}'"
                )

            # apply to mask
            mask_datums = np.isin(self._detailed_pairs[:, 0], datums)

        filtered_detailed_pairs = self._detailed_pairs[mask_datums]
        n_pairs = self._detailed_pairs[mask_datums].shape[0]
        mask_groundtruths = np.zeros(n_pairs, dtype=np.bool_)
        mask_predictions = np.zeros_like(mask_groundtruths)

        # filter by ground truth annotation ids
        if groundtruths is not None:
            # convert to indices
            if isinstance(groundtruths, list):
                groundtruths = np.array(
                    [
                        self.groundtruth_id_to_index[uid]
                        for uid in groundtruths
                    ],
                    dtype=np.int32,
                )

            # validate indices
            if groundtruths.size == 0:
                warnings.warn("filter removes all ground truths")
            elif groundtruths.min() < 0:
                raise ValueError(
                    f"groundtruth annotation index cannot be negative '{groundtruths.min()}'"
                )
            elif groundtruths.max() >= len(self.index_to_groundtruth_id):
                raise ValueError(
                    f"groundtruth annotation index cannot exceed total number of groundtruths '{groundtruths.max()}'"
                )

            # apply to mask
            mask_groundtruths[
                ~np.isin(
                    filtered_detailed_pairs[:, 1],
                    groundtruths,
                )
            ] = True

        # filter by prediction annotation ids
        if predictions is not None:
            # convert to indices
            if isinstance(predictions, list):
                predictions = np.array(
                    [self.prediction_id_to_index[uid] for uid in predictions],
                    dtype=np.int32,
                )

            # validate indices
            if predictions.size == 0:
                warnings.warn("filter removes all predictions")
            elif predictions.min() < 0:
                raise ValueError(
                    f"prediction annotation index cannot be negative '{predictions.min()}'"
                )
            elif predictions.max() >= len(self.index_to_prediction_id):
                raise ValueError(
                    f"prediction annotation index cannot exceed total number of predictions '{predictions.max()}'"
                )

            # apply to mask
            mask_predictions[
                ~np.isin(
                    filtered_detailed_pairs[:, 2],
                    predictions,
                )
            ] = True

        # filter by labels
        if labels is not None:
            # convert to indices
            if isinstance(labels, list):
                labels = np.array(
                    [self.label_to_index[label] for label in labels]
                )

            # validate indices
            if labels.size == 0:
                raise EmptyFilterError("filter removes all labels")
            elif labels.min() < 0:
                raise ValueError(
                    f"label index cannot be negative '{labels.min()}'"
                )
            elif labels.max() >= len(self.index_to_label):
                raise ValueError(
                    f"label index cannot exceed total number of labels '{labels.max()}'"
                )

            # apply to mask
            labels = np.concatenate([labels, np.array([-1])])  # add null label
            mask_groundtruths[
                ~np.isin(filtered_detailed_pairs[:, 3], labels)
            ] = True
            mask_predictions[
                ~np.isin(filtered_detailed_pairs[:, 4], labels)
            ] = True

        filtered_detailed_pairs, _, _ = filter_cache(
            self._detailed_pairs,
            mask_datums=mask_datums,
            mask_ground_truths=mask_groundtruths,
            mask_predictions=mask_predictions,
            n_labels=len(self.index_to_label),
        )

        number_of_datums = (
            datums.size
            if datums is not None
            else np.unique(filtered_detailed_pairs[:, 0]).size
        )

        return Filter(
            mask_datums=mask_datums,
            mask_groundtruths=mask_groundtruths,
            mask_predictions=mask_predictions,
            metadata=Metadata.create(
                detailed_pairs=filtered_detailed_pairs,
                number_of_datums=number_of_datums,
                number_of_labels=len(self.index_to_label),
            ),
        )

    def filter(
        self, filter_: Filter
    ) -> tuple[NDArray[np.float64], NDArray[np.float64], NDArray[np.int32],]:
        """
        Performs filtering over the internal cache.

        Parameters
        ----------
        filter_ : Filter
            The filter parameterization.

        Returns
        -------
        NDArray[float64]
            Filtered detailed pairs.
        NDArray[float64]
            Filtered ranked pairs.
        NDArray[int32]
            Label metadata.
        """
        return filter_cache(
            detailed_pairs=self._detailed_pairs,
            mask_datums=filter_.mask_datums,
            mask_ground_truths=filter_.mask_groundtruths,
            mask_predictions=filter_.mask_predictions,
            n_labels=len(self.index_to_label),
        )

    def compute_precision_recall(
        self,
        iou_thresholds: list[float],
        score_thresholds: list[float],
        filter_: Filter | None = None,
    ) -> dict[MetricType, list[Metric]]:
        """
        Computes all metrics except for ConfusionMatrix

        Parameters
        ----------
        iou_thresholds : list[float]
            A list of IOU thresholds to compute metrics over.
        score_thresholds : list[float]
            A list of score thresholds to compute metrics over.
        filter_ : Filter, optional
            A collection of filter parameters and masks.

        Returns
        -------
        dict[MetricType, list]
            A dictionary mapping MetricType enumerations to lists of computed metrics.
        """
        if not iou_thresholds:
            raise ValueError("At least one IOU threshold must be passed.")
        elif not score_thresholds:
            raise ValueError("At least one score threshold must be passed.")

        if filter_ is not None:
            _, ranked_pairs, label_metadata = self.filter(filter_=filter_)
        else:
            ranked_pairs = self._ranked_pairs
            label_metadata = self._label_metadata

        results = compute_precion_recall(
            ranked_pairs=ranked_pairs,
            label_metadata=label_metadata,
            iou_thresholds=np.array(iou_thresholds),
            score_thresholds=np.array(score_thresholds),
        )
        return unpack_precision_recall_into_metric_lists(
            results=results,
            label_metadata=label_metadata,
            iou_thresholds=iou_thresholds,
            score_thresholds=score_thresholds,
            index_to_label=self.index_to_label,
        )

    def compute_confusion_matrix(
        self,
        iou_thresholds: list[float],
        score_thresholds: list[float],
        filter_: Filter | None = None,
    ) -> list[Metric]:
        """
        Computes confusion matrices at various thresholds.

        Parameters
        ----------
        iou_thresholds : list[float]
            A list of IOU thresholds to compute metrics over.
        score_thresholds : list[float]
            A list of score thresholds to compute metrics over.
        filter_ : Filter, optional
            A collection of filter parameters and masks.

        Returns
        -------
        list[Metric]
            List of confusion matrices per threshold pair.
        """
        if not iou_thresholds:
            raise ValueError("At least one IOU threshold must be passed.")
        elif not score_thresholds:
            raise ValueError("At least one score threshold must be passed.")

        if filter_ is not None:
            detailed_pairs, _, _ = self.filter(filter_=filter_)
        else:
            detailed_pairs = self._detailed_pairs

        if detailed_pairs.size == 0:
            return []

        results = compute_confusion_matrix(
            detailed_pairs=detailed_pairs,
            iou_thresholds=np.array(iou_thresholds),
            score_thresholds=np.array(score_thresholds),
        )
        return unpack_confusion_matrix_into_metric_list(
            results=results,
            detailed_pairs=detailed_pairs,
            iou_thresholds=iou_thresholds,
            score_thresholds=score_thresholds,
            index_to_datum_id=self.index_to_datum_id,
            index_to_groundtruth_id=self.index_to_groundtruth_id,
            index_to_prediction_id=self.index_to_prediction_id,
            index_to_label=self.index_to_label,
        )

    def evaluate(
        self,
        iou_thresholds: list[float] = [0.1, 0.5, 0.75],
        score_thresholds: list[float] = [0.5],
        filter_: Filter | None = None,
    ) -> dict[MetricType, list[Metric]]:
        """
        Computes all available metrics.

        Parameters
        ----------
        iou_thresholds : list[float], default=[0.1, 0.5, 0.75]
            A list of IOU thresholds to compute metrics over.
        score_thresholds : list[float], default=[0.5]
            A list of score thresholds to compute metrics over.
        filter_ : Filter, optional
            A collection of filter parameters and masks.

        Returns
        -------
        dict[MetricType, list[Metric]]
            Lists of metrics organized by metric type.
        """
        metrics = self.compute_precision_recall(
            iou_thresholds=iou_thresholds,
            score_thresholds=score_thresholds,
            filter_=filter_,
        )
        metrics[MetricType.ConfusionMatrix] = self.compute_confusion_matrix(
            iou_thresholds=iou_thresholds,
            score_thresholds=score_thresholds,
            filter_=filter_,
        )
        return metrics


class DataLoader:
    """
    Object Detection DataLoader
    """

    def __init__(self):
        self._evaluator = Evaluator()
        self.pairs: list[NDArray[np.float64]] = list()

    def _add_datum(self, datum_id: str) -> int:
        """
        Helper function for adding a datum to the cache.

        Parameters
        ----------
        datum_id : str
            The datum identifier.

        Returns
        -------
        int
            The datum index.
        """
        if datum_id not in self._evaluator.datum_id_to_index:
            if len(self._evaluator.datum_id_to_index) != len(
                self._evaluator.index_to_datum_id
            ):
                raise InternalCacheError("datum cache size mismatch")
            idx = len(self._evaluator.datum_id_to_index)
            self._evaluator.datum_id_to_index[datum_id] = idx
            self._evaluator.index_to_datum_id.append(datum_id)
        return self._evaluator.datum_id_to_index[datum_id]

    def _add_groundtruth(self, annotation_id: str) -> int:
        """
        Helper function for adding a ground truth annotation identifier to the cache.

        Parameters
        ----------
        annotation_id : str
            The ground truth annotation identifier.

        Returns
        -------
        int
            The ground truth annotation index.
        """
        if annotation_id not in self._evaluator.groundtruth_id_to_index:
            if len(self._evaluator.groundtruth_id_to_index) != len(
                self._evaluator.index_to_groundtruth_id
            ):
                raise InternalCacheError("ground truth cache size mismatch")
            idx = len(self._evaluator.groundtruth_id_to_index)
            self._evaluator.groundtruth_id_to_index[annotation_id] = idx
            self._evaluator.index_to_groundtruth_id.append(annotation_id)
        return self._evaluator.groundtruth_id_to_index[annotation_id]

    def _add_prediction(self, annotation_id: str) -> int:
        """
        Helper function for adding a prediction annotation identifier to the cache.

        Parameters
        ----------
        annotation_id : str
            The prediction annotation identifier.

        Returns
        -------
        int
            The prediction annotation index.
        """
        if annotation_id not in self._evaluator.prediction_id_to_index:
            if len(self._evaluator.prediction_id_to_index) != len(
                self._evaluator.index_to_prediction_id
            ):
                raise InternalCacheError("prediction cache size mismatch")
            idx = len(self._evaluator.prediction_id_to_index)
            self._evaluator.prediction_id_to_index[annotation_id] = idx
            self._evaluator.index_to_prediction_id.append(annotation_id)
        return self._evaluator.prediction_id_to_index[annotation_id]

    def _add_label(self, label: str) -> int:
        """
        Helper function for adding a label to the cache.

        Parameters
        ----------
        label : str
            The label associated with the annotation.

        Returns
        -------
        int
            Label index.
        """
        label_id = len(self._evaluator.index_to_label)
        if label not in self._evaluator.label_to_index:
            if len(self._evaluator.label_to_index) != len(
                self._evaluator.index_to_label
            ):
                raise InternalCacheError("label cache size mismatch")
            self._evaluator.label_to_index[label] = label_id
            self._evaluator.index_to_label.append(label)
            label_id += 1
        return self._evaluator.label_to_index[label]

    def _add_data(
        self,
        detections: list[Detection],
        detection_ious: list[NDArray[np.float64]],
        show_progress: bool = False,
    ):
        """
        Adds detections to the cache.

        Parameters
        ----------
        detections : list[Detection]
            A list of Detection objects.
        detection_ious : list[NDArray[np.float64]]
            A list of arrays containing IOUs per detection.
        show_progress : bool, default=False
            Toggle for tqdm progress bar.
        """
        disable_tqdm = not show_progress
        for detection, ious in tqdm(
            zip(detections, detection_ious), disable=disable_tqdm
        ):
            # cache labels and annotation pairs
            pairs = []
            datum_idx = self._add_datum(detection.uid)
            if detection.groundtruths:
                for gidx, gann in enumerate(detection.groundtruths):
                    groundtruth_idx = self._add_groundtruth(gann.uid)
                    glabel_idx = self._add_label(gann.labels[0])
                    if (ious[:, gidx] < 1e-9).all():
                        pairs.extend(
                            [
                                np.array(
                                    [
                                        float(datum_idx),
                                        float(groundtruth_idx),
                                        -1.0,
                                        float(glabel_idx),
                                        -1.0,
                                        0.0,
                                        -1.0,
                                    ]
                                )
                            ]
                        )
                    for pidx, pann in enumerate(detection.predictions):
                        prediction_idx = self._add_prediction(pann.uid)
                        if (ious[pidx, :] < 1e-9).all():
                            pairs.extend(
                                [
                                    np.array(
                                        [
                                            float(datum_idx),
                                            -1.0,
                                            float(prediction_idx),
                                            -1.0,
                                            float(self._add_label(plabel)),
                                            0.0,
                                            float(pscore),
                                        ]
                                    )
                                    for plabel, pscore in zip(
                                        pann.labels, pann.scores
                                    )
                                ]
                            )
                        if ious[pidx, gidx] >= 1e-9:
                            pairs.extend(
                                [
                                    np.array(
                                        [
                                            float(datum_idx),
                                            float(groundtruth_idx),
                                            float(prediction_idx),
                                            float(self._add_label(glabel)),
                                            float(self._add_label(plabel)),
                                            ious[pidx, gidx],
                                            float(pscore),
                                        ]
                                    )
                                    for glabel in gann.labels
                                    for plabel, pscore in zip(
                                        pann.labels, pann.scores
                                    )
                                ]
                            )
            elif detection.predictions:
                for pidx, pann in enumerate(detection.predictions):
                    prediction_idx = self._add_prediction(pann.uid)
                    pairs.extend(
                        [
                            np.array(
                                [
                                    float(datum_idx),
                                    -1.0,
                                    float(prediction_idx),
                                    -1.0,
                                    float(self._add_label(plabel)),
                                    0.0,
                                    float(pscore),
                                ]
                            )
                            for plabel, pscore in zip(pann.labels, pann.scores)
                        ]
                    )

            data = np.array(pairs)
            if data.size > 0:
                self.pairs.append(data)

    def add_bounding_boxes(
        self,
        detections: list[Detection[BoundingBox]],
        show_progress: bool = False,
    ):
        """
        Adds bounding box detections to the cache.

        Parameters
        ----------
        detections : list[Detection]
            A list of Detection objects.
        show_progress : bool, default=False
            Toggle for tqdm progress bar.
        """
        ious = [
            compute_bbox_iou(
                np.array(
                    [
                        [gt.extrema, pd.extrema]
                        for pd in detection.predictions
                        for gt in detection.groundtruths
                    ],
                    dtype=np.float64,
                )
            ).reshape(len(detection.predictions), len(detection.groundtruths))
            for detection in detections
        ]
        return self._add_data(
            detections=detections,
            detection_ious=ious,
            show_progress=show_progress,
        )

    def add_polygons(
        self,
        detections: list[Detection[Polygon]],
        show_progress: bool = False,
    ):
        """
        Adds polygon detections to the cache.

        Parameters
        ----------
        detections : list[Detection]
            A list of Detection objects.
        show_progress : bool, default=False
            Toggle for tqdm progress bar.
        """
        ious = [
            compute_polygon_iou(
                np.array(
                    [
                        [gt.shape, pd.shape]
                        for pd in detection.predictions
                        for gt in detection.groundtruths
                    ]
                )
            ).reshape(len(detection.predictions), len(detection.groundtruths))
            for detection in detections
        ]
        return self._add_data(
            detections=detections,
            detection_ious=ious,
            show_progress=show_progress,
        )

    def add_bitmasks(
        self,
        detections: list[Detection[Bitmask]],
        show_progress: bool = False,
    ):
        """
        Adds bitmask detections to the cache.

        Parameters
        ----------
        detections : list[Detection]
            A list of Detection objects.
        show_progress : bool, default=False
            Toggle for tqdm progress bar.
        """
        ious = [
            compute_bitmask_iou(
                np.array(
                    [
                        [gt.mask, pd.mask]
                        for pd in detection.predictions
                        for gt in detection.groundtruths
                    ]
                )
            ).reshape(len(detection.predictions), len(detection.groundtruths))
            for detection in detections
        ]
        return self._add_data(
            detections=detections,
            detection_ious=ious,
            show_progress=show_progress,
        )

    def finalize(self) -> Evaluator:
        """
        Performs data finalization and some preprocessing steps.

        Returns
        -------
        Evaluator
            A ready-to-use evaluator object.
        """
        if not self.pairs:
            raise EmptyEvaluatorError()

        n_labels = len(self._evaluator.index_to_label)
        n_datums = len(self._evaluator.index_to_datum_id)

        self._evaluator._detailed_pairs = np.concatenate(self.pairs, axis=0)
        if self._evaluator._detailed_pairs.size == 0:
            raise EmptyEvaluatorError()

        # order pairs by descending score, iou
        indices = np.lexsort(
            (
                -self._evaluator._detailed_pairs[:, 5],  # iou
                -self._evaluator._detailed_pairs[:, 6],  # score
            )
        )
        self._evaluator._detailed_pairs = self._evaluator._detailed_pairs[
            indices
        ]
        self._evaluator._label_metadata = compute_label_metadata(
            ids=self._evaluator._detailed_pairs[:, :5].astype(np.int32),
            n_labels=n_labels,
        )
        self._evaluator._ranked_pairs = rank_pairs(
            detailed_pairs=self._evaluator._detailed_pairs,
            label_metadata=self._evaluator._label_metadata,
        )
        self._evaluator._metadata = Metadata.create(
            detailed_pairs=self._evaluator._detailed_pairs,
            number_of_datums=n_datums,
            number_of_labels=n_labels,
        )
        return self._evaluator



from enum import IntFlag, auto

import numpy as np
from numpy.typing import NDArray

import valor_lite.classification.numpy_compatibility as npc


def compute_label_metadata(
    ids: NDArray[np.int32],
    n_labels: int,
) -> NDArray[np.int32]:
    """
    Computes label metadata returning a count of annotations per label.

    Parameters
    ----------
    detailed_pairs : NDArray[np.int32]
        Detailed annotation pairings with shape (n_pairs, 3).
            Index 0 - Datum Index
            Index 1 - GroundTruth Label Index
            Index 2 - Prediction Label Index
    n_labels : int
        The total number of unique labels.

    Returns
    -------
    NDArray[np.int32]
        The label metadata array with shape (n_labels, 2).
            Index 0 - Ground truth label count
            Index 1 - Prediction label count
    """
    label_metadata = np.zeros((n_labels, 2), dtype=np.int32)
    ground_truth_pairs = ids[:, (0, 1)]
    ground_truth_pairs = ground_truth_pairs[ground_truth_pairs[:, 1] >= 0]
    unique_pairs = np.unique(ground_truth_pairs, axis=0)
    label_indices, unique_counts = np.unique(
        unique_pairs[:, 1], return_counts=True
    )
    label_metadata[label_indices.astype(np.int32), 0] = unique_counts

    prediction_pairs = ids[:, (0, 2)]
    prediction_pairs = prediction_pairs[prediction_pairs[:, 1] >= 0]
    unique_pairs = np.unique(prediction_pairs, axis=0)
    label_indices, unique_counts = np.unique(
        unique_pairs[:, 1], return_counts=True
    )
    label_metadata[label_indices.astype(np.int32), 1] = unique_counts

    return label_metadata


def filter_cache(
    detailed_pairs: NDArray[np.float64],
    datum_mask: NDArray[np.bool_],
    valid_label_indices: NDArray[np.int32] | None,
    n_labels: int,
) -> tuple[NDArray[np.float64], NDArray[np.int32]]:
    # filter by datum
    detailed_pairs = detailed_pairs[datum_mask].copy()

    n_rows = detailed_pairs.shape[0]
    mask_invalid_groundtruths = np.zeros(n_rows, dtype=np.bool_)
    mask_invalid_predictions = np.zeros_like(mask_invalid_groundtruths)

    # filter labels
    if valid_label_indices is not None:
        mask_invalid_groundtruths[
            ~np.isin(detailed_pairs[:, 1], valid_label_indices)
        ] = True
        mask_invalid_predictions[
            ~np.isin(detailed_pairs[:, 2], valid_label_indices)
        ] = True

    # filter cache
    if mask_invalid_groundtruths.any():
        invalid_groundtruth_indices = np.where(mask_invalid_groundtruths)[0]
        detailed_pairs[invalid_groundtruth_indices[:, None], 1] = np.array(
            [[-1.0]]
        )

    if mask_invalid_predictions.any():
        invalid_prediction_indices = np.where(mask_invalid_predictions)[0]
        detailed_pairs[
            invalid_prediction_indices[:, None], (2, 3, 4)
        ] = np.array([[-1.0, -1.0, -1.0]])

    # filter null pairs
    mask_null_pairs = np.all(
        np.isclose(
            detailed_pairs[:, 1:5],
            np.array([-1.0, -1.0, -1.0, -1.0]),
        ),
        axis=1,
    )
    detailed_pairs = detailed_pairs[~mask_null_pairs]

    detailed_pairs = np.unique(detailed_pairs, axis=0)
    indices = np.lexsort(
        (
            detailed_pairs[:, 1],  # ground truth
            detailed_pairs[:, 2],  # prediction
            -detailed_pairs[:, 3],  # score
        )
    )
    detailed_pairs = detailed_pairs[indices]
    label_metadata = compute_label_metadata(
        ids=detailed_pairs[:, :3].astype(np.int32),
        n_labels=n_labels,
    )
    return detailed_pairs, label_metadata


def _compute_rocauc(
    data: NDArray[np.float64],
    label_metadata: NDArray[np.int32],
    n_datums: int,
    n_labels: int,
    mask_matching_labels: NDArray[np.bool_],
    pd_labels: NDArray[np.int32],
) -> tuple[NDArray[np.float64], float]:
    """
    Compute ROCAUC and mean ROCAUC.
    """
    positive_count = label_metadata[:, 0]
    negative_count = label_metadata[:, 1] - label_metadata[:, 0]

    true_positives = np.zeros((n_labels, n_datums), dtype=np.int32)
    false_positives = np.zeros_like(true_positives)
    scores = np.zeros_like(true_positives, dtype=np.float64)

    for label_idx in range(n_labels):
        if label_metadata[label_idx, 1] == 0:
            continue

        mask_pds = pd_labels == label_idx
        true_positives[label_idx] = mask_matching_labels[mask_pds]
        false_positives[label_idx] = ~mask_matching_labels[mask_pds]
        scores[label_idx] = data[mask_pds, 3]

    cumulative_fp = np.cumsum(false_positives, axis=1)
    cumulative_tp = np.cumsum(true_positives, axis=1)

    fpr = np.zeros_like(true_positives, dtype=np.float64)
    np.divide(
        cumulative_fp,
        negative_count[:, np.newaxis],
        where=negative_count[:, np.newaxis] > 1e-9,
        out=fpr,
    )
    tpr = np.zeros_like(true_positives, dtype=np.float64)
    np.divide(
        cumulative_tp,
        positive_count[:, np.newaxis],
        where=positive_count[:, np.newaxis] > 1e-9,
        out=tpr,
    )

    # sort by -tpr, -score
    indices = np.lexsort((-tpr, -scores), axis=1)
    fpr = np.take_along_axis(fpr, indices, axis=1)
    tpr = np.take_along_axis(tpr, indices, axis=1)

    # running max of tpr
    np.maximum.accumulate(tpr, axis=1, out=tpr)

    # compute rocauc
    rocauc = npc.trapezoid(x=fpr, y=tpr, axis=1)

    # compute mean rocauc
    mean_rocauc = rocauc.mean()

    return rocauc, mean_rocauc  # type: ignore[reportReturnType]


def compute_precision_recall_rocauc(
    detailed_pairs: NDArray[np.float64],
    label_metadata: NDArray[np.int32],
    score_thresholds: NDArray[np.float64],
    hardmax: bool,
    n_datums: int,
) -> tuple[
    NDArray[np.int32],
    NDArray[np.float64],
    NDArray[np.float64],
    NDArray[np.float64],
    NDArray[np.float64],
    NDArray[np.float64],
    float,
]:
    """
    Computes classification metrics.

    Parameters
    ----------
    detailed_pairs : NDArray[np.float64]
        A sorted array of classification pairs with shape (n_pairs, 5).
            Index 0 - Datum Index
            Index 1 - GroundTruth Label Index
            Index 2 - Prediction Label Index
            Index 3 - Score
            Index 4 - Hard-Max Score
    label_metadata : NDArray[np.int32]
        An array containing metadata related to labels with shape (n_labels, 2).
            Index 0 - GroundTruth Label Count
            Index 1 - Prediction Label Count
    score_thresholds : NDArray[np.float64]
        A 1-D array contains score thresholds to compute metrics over.
    hardmax : bool
        Option to only allow a single positive prediction.
    n_datums : int
        The number of datums being operated over.

    Returns
    -------
    NDArray[np.int32]
        TP, FP, FN, TN counts.
    NDArray[np.float64]
        Precision.
    NDArray[np.float64]
        Recall.
    NDArray[np.float64]
        Accuracy
    NDArray[np.float64]
        F1 Score
    NDArray[np.float64]
        ROCAUC.
    float
        mROCAUC.
    """

    n_labels = label_metadata.shape[0]
    n_scores = score_thresholds.shape[0]

    pd_labels = detailed_pairs[:, 2].astype(int)

    mask_matching_labels = np.isclose(
        detailed_pairs[:, 1], detailed_pairs[:, 2]
    )
    mask_score_nonzero = ~np.isclose(detailed_pairs[:, 3], 0.0)
    mask_hardmax = detailed_pairs[:, 4] > 0.5

    # calculate ROCAUC
    rocauc, mean_rocauc = _compute_rocauc(
        data=detailed_pairs,
        label_metadata=label_metadata,
        n_datums=n_datums,
        n_labels=n_labels,
        mask_matching_labels=mask_matching_labels,
        pd_labels=pd_labels,
    )

    # calculate metrics at various score thresholds
    counts = np.zeros((n_scores, n_labels, 4), dtype=np.int32)
    for score_idx in range(n_scores):
        mask_score_threshold = (
            detailed_pairs[:, 3] >= score_thresholds[score_idx]
        )
        mask_score = mask_score_nonzero & mask_score_threshold

        if hardmax:
            mask_score &= mask_hardmax

        mask_tp = mask_matching_labels & mask_score
        mask_fp = ~mask_matching_labels & mask_score
        mask_fn = (mask_matching_labels & ~mask_score) | mask_fp
        mask_tn = ~mask_matching_labels & ~mask_score

        fn = np.unique(detailed_pairs[mask_fn][:, [0, 1]].astype(int), axis=0)
        tn = np.unique(detailed_pairs[mask_tn][:, [0, 2]].astype(int), axis=0)

        counts[score_idx, :, 0] = np.bincount(
            pd_labels[mask_tp], minlength=n_labels
        )
        counts[score_idx, :, 1] = np.bincount(
            pd_labels[mask_fp], minlength=n_labels
        )
        counts[score_idx, :, 2] = np.bincount(fn[:, 1], minlength=n_labels)
        counts[score_idx, :, 3] = np.bincount(tn[:, 1], minlength=n_labels)

    recall = np.zeros((n_scores, n_labels), dtype=np.float64)
    np.divide(
        counts[:, :, 0],
        (counts[:, :, 0] + counts[:, :, 2]),
        where=(counts[:, :, 0] + counts[:, :, 2]) > 1e-9,
        out=recall,
    )

    precision = np.zeros_like(recall)
    np.divide(
        counts[:, :, 0],
        (counts[:, :, 0] + counts[:, :, 1]),
        where=(counts[:, :, 0] + counts[:, :, 1]) > 1e-9,
        out=precision,
    )

    accuracy = np.zeros(n_scores, dtype=np.float64)
    np.divide(
        counts[:, :, 0].sum(axis=1),
        float(n_datums),
        out=accuracy,
    )

    f1_score = np.zeros_like(recall)
    np.divide(
        (2 * precision * recall),
        (precision + recall),
        where=(precision + recall) > 1e-9,
        out=f1_score,
    )

    return (
        counts,
        precision,
        recall,
        accuracy,
        f1_score,
        rocauc,
        mean_rocauc,
    )


class PairClassification(IntFlag):
    TP = auto()
    FP_FN_MISCLF = auto()
    FN_UNMATCHED = auto()


def compute_confusion_matrix(
    detailed_pairs: NDArray[np.float64],
    score_thresholds: NDArray[np.float64],
    hardmax: bool,
) -> NDArray[np.uint8]:
    """
    Compute detailed confusion matrix.

    Parameters
    ----------
    detailed_pairs : NDArray[np.float64]
        A 2-D sorted array summarizing the IOU calculations of one or more pairs with shape (n_pairs, 5).
            Index 0 - Datum Index
            Index 1 - GroundTruth Label Index
            Index 2 - Prediction Label Index
            Index 3 - Score
            Index 4 - Hard Max Score
    iou_thresholds : NDArray[np.float64]
        A 1-D array containing IOU thresholds.
    score_thresholds : NDArray[np.float64]
        A 1-D array containing score thresholds.

    Returns
    -------
    NDArray[uint8]
        Row-wise classification of pairs.
    """
    n_pairs = detailed_pairs.shape[0]
    n_scores = score_thresholds.shape[0]

    pair_classifications = np.zeros(
        (n_scores, n_pairs),
        dtype=np.uint8,
    )

    mask_label_match = np.isclose(detailed_pairs[:, 1], detailed_pairs[:, 2])
    mask_score = detailed_pairs[:, 3] > 1e-9

    groundtruths = detailed_pairs[:, [0, 1]].astype(int)

    for score_idx in range(n_scores):
        mask_score &= detailed_pairs[:, 3] >= score_thresholds[score_idx]
        if hardmax:
            mask_score &= detailed_pairs[:, 4] > 0.5

        mask_true_positives = mask_label_match & mask_score
        mask_misclassifications = ~mask_label_match & mask_score
        mask_unmatched_groundtruths = ~(
            (
                groundtruths.reshape(-1, 1, 2)
                == groundtruths[mask_score].reshape(1, -1, 2)
            )
            .all(axis=2)
            .any(axis=1)
        )

        # classify pairings
        pair_classifications[score_idx, mask_true_positives] |= np.uint8(
            PairClassification.TP
        )
        pair_classifications[score_idx, mask_misclassifications] |= np.uint8(
            PairClassification.FP_FN_MISCLF
        )
        pair_classifications[
            score_idx, mask_unmatched_groundtruths
        ] |= np.uint8(PairClassification.FN_UNMATCHED)

    return pair_classifications



from enum import IntFlag, auto

import numpy as np
import shapely
from numpy.typing import NDArray


def compute_bbox_iou(data: NDArray[np.float64]) -> NDArray[np.float64]:
    """
    Computes intersection-over-union (IOU) for axis-aligned bounding boxes.

    Takes data with shape (N, 8):

    Index 0 - xmin for Box 1
    Index 1 - xmax for Box 1
    Index 2 - ymin for Box 1
    Index 3 - ymax for Box 1
    Index 4 - xmin for Box 2
    Index 5 - xmax for Box 2
    Index 6 - ymin for Box 2
    Index 7 - ymax for Box 2

    Returns data with shape (N, 1):

    Index 0 - IOU

    Parameters
    ----------
    data : NDArray[np.float64]
        A sorted array of bounding box pairs.

    Returns
    -------
    NDArray[np.float64]
        Computed IOU's.
    """
    if data.size == 0:
        return np.array([], dtype=np.float64)

    n_pairs = data.shape[0]

    xmin1, xmax1, ymin1, ymax1 = (
        data[:, 0, 0],
        data[:, 0, 1],
        data[:, 0, 2],
        data[:, 0, 3],
    )
    xmin2, xmax2, ymin2, ymax2 = (
        data[:, 1, 0],
        data[:, 1, 1],
        data[:, 1, 2],
        data[:, 1, 3],
    )

    xmin = np.maximum(xmin1, xmin2)
    ymin = np.maximum(ymin1, ymin2)
    xmax = np.minimum(xmax1, xmax2)
    ymax = np.minimum(ymax1, ymax2)

    intersection_width = np.maximum(0, xmax - xmin)
    intersection_height = np.maximum(0, ymax - ymin)
    intersection_area = intersection_width * intersection_height

    area1 = (xmax1 - xmin1) * (ymax1 - ymin1)
    area2 = (xmax2 - xmin2) * (ymax2 - ymin2)

    union_area = area1 + area2 - intersection_area

    ious = np.zeros(n_pairs, dtype=np.float64)
    np.divide(
        intersection_area,
        union_area,
        where=union_area >= 1e-9,
        out=ious,
    )
    return ious


def compute_bitmask_iou(data: NDArray[np.bool_]) -> NDArray[np.float64]:
    """
    Computes intersection-over-union (IOU) for bitmasks.

    Takes data with shape (N, 2):

    Index 0 - first bitmask
    Index 1 - second bitmask

    Returns data with shape (N, 1):

    Index 0 - IOU

    Parameters
    ----------
    data : NDArray[np.float64]
        A sorted array of bitmask pairs.

    Returns
    -------
    NDArray[np.float64]
        Computed IOU's.
    """

    if data.size == 0:
        return np.array([], dtype=np.float64)

    n_pairs = data.shape[0]
    lhs = data[:, 0, :, :].reshape(n_pairs, -1)
    rhs = data[:, 1, :, :].reshape(n_pairs, -1)

    lhs_sum = lhs.sum(axis=1)
    rhs_sum = rhs.sum(axis=1)

    intersection_ = np.logical_and(lhs, rhs).sum(axis=1)
    union_ = lhs_sum + rhs_sum - intersection_

    ious = np.zeros(n_pairs, dtype=np.float64)
    np.divide(
        intersection_,
        union_,
        where=union_ >= 1e-9,
        out=ious,
    )
    return ious


def compute_polygon_iou(
    data: NDArray[np.float64],
) -> NDArray[np.float64]:
    """
    Computes intersection-over-union (IOU) for shapely polygons.

    Takes data with shape (N, 2):

    Index 0 - first polygon
    Index 1 - second polygon

    Returns data with shape (N, 1):

    Index 0 - IOU

    Parameters
    ----------
    data : NDArray[np.float64]
        A sorted array of polygon pairs.

    Returns
    -------
    NDArray[np.float64]
        Computed IOU's.
    """

    if data.size == 0:
        return np.array([], dtype=np.float64)

    n_pairs = data.shape[0]

    lhs = data[:, 0]
    rhs = data[:, 1]

    intersections = shapely.intersection(lhs, rhs)
    intersection_areas = shapely.area(intersections)

    unions = shapely.union(lhs, rhs)
    union_areas = shapely.area(unions)

    ious = np.zeros(n_pairs, dtype=np.float64)
    np.divide(
        intersection_areas,
        union_areas,
        where=union_areas >= 1e-9,
        out=ious,
    )
    return ious


def compute_label_metadata(
    ids: NDArray[np.int32],
    n_labels: int,
) -> NDArray[np.int32]:
    """
    Computes label metadata returning a count of annotations per label.

    Parameters
    ----------
    detailed_pairs : NDArray[np.int32]
        Detailed annotation pairings with shape (N, 7).
            Index 0 - Datum Index
            Index 1 - GroundTruth Index
            Index 2 - Prediction Index
            Index 3 - GroundTruth Label Index
            Index 4 - Prediction Label Index
    n_labels : int
        The total number of unique labels.

    Returns
    -------
    NDArray[np.int32]
        The label metadata array with shape (n_labels, 2).
            Index 0 - Ground truth label count
            Index 1 - Prediction label count
    """
    label_metadata = np.zeros((n_labels, 2), dtype=np.int32)

    ground_truth_pairs = ids[:, (0, 1, 3)]
    ground_truth_pairs = ground_truth_pairs[ground_truth_pairs[:, 1] >= 0]
    unique_pairs = np.unique(ground_truth_pairs, axis=0)
    label_indices, unique_counts = np.unique(
        unique_pairs[:, 2], return_counts=True
    )
    label_metadata[label_indices.astype(np.int32), 0] = unique_counts

    prediction_pairs = ids[:, (0, 2, 4)]
    prediction_pairs = prediction_pairs[prediction_pairs[:, 1] >= 0]
    unique_pairs = np.unique(prediction_pairs, axis=0)
    label_indices, unique_counts = np.unique(
        unique_pairs[:, 2], return_counts=True
    )
    label_metadata[label_indices.astype(np.int32), 1] = unique_counts

    return label_metadata


def filter_cache(
    detailed_pairs: NDArray[np.float64],
    mask_datums: NDArray[np.bool_],
    mask_predictions: NDArray[np.bool_],
    mask_ground_truths: NDArray[np.bool_],
    n_labels: int,
) -> tuple[NDArray[np.float64], NDArray[np.float64], NDArray[np.int32],]:
    """
    Performs filtering on a detailed cache.

    Parameters
    ----------
    detailed_pairs : NDArray[float64]
        A list of sorted detailed pairs with size (N, 7).
    mask_datums : NDArray[bool]
        A boolean mask with size (N,).
    mask_ground_truths : NDArray[bool]
        A boolean mask with size (N,).
    mask_predictions : NDArray[bool]
        A boolean mask with size (N,).
    n_labels : int
        The total number of unique labels.

    Returns
    -------
    NDArray[float64]
        Filtered detailed pairs.
    NDArray[float64]
        Filtered ranked pairs.
    NDArray[int32]
        Label metadata.
    """
    # filter datums
    detailed_pairs = detailed_pairs[mask_datums].copy()

    # filter ground truths
    if mask_ground_truths.any():
        invalid_groundtruth_indices = np.where(mask_ground_truths)[0]
        detailed_pairs[
            invalid_groundtruth_indices[:, None], (1, 3, 5)
        ] = np.array([[-1, -1, 0]])

    # filter predictions
    if mask_predictions.any():
        invalid_prediction_indices = np.where(mask_predictions)[0]
        detailed_pairs[
            invalid_prediction_indices[:, None], (2, 4, 5, 6)
        ] = np.array([[-1, -1, 0, -1]])

    # filter null pairs
    mask_null_pairs = np.all(
        np.isclose(
            detailed_pairs[:, 1:5],
            np.array([-1.0, -1.0, -1.0, -1.0]),
        ),
        axis=1,
    )
    detailed_pairs = detailed_pairs[~mask_null_pairs]

    # sorts by score, iou with ground truth id as a tie-breaker
    indices = np.lexsort(
        (
            detailed_pairs[:, 1],  # ground truth id
            -detailed_pairs[:, 5],  # iou
            -detailed_pairs[:, 6],  # score
        )
    )
    detailed_pairs = detailed_pairs[indices]
    label_metadata = compute_label_metadata(
        ids=detailed_pairs[:, :5].astype(np.int32),
        n_labels=n_labels,
    )
    ranked_pairs = rank_pairs(
        detailed_pairs=detailed_pairs,
        label_metadata=label_metadata,
    )
    return (
        detailed_pairs,
        ranked_pairs,
        label_metadata,
    )


def rank_pairs(
    detailed_pairs: NDArray[np.float64],
    label_metadata: NDArray[np.int32],
) -> NDArray[np.float64]:
    """
    Highly optimized pair ranking for computing precision and recall based metrics.

    Only ground truths and predictions that provide unique information are kept. The unkept
    pairs are represented via the label metadata array.

    Parameters
    ----------
    detailed_pairs : NDArray[np.float64]
        Detailed annotation pairs with shape (n_pairs, 7).
            Index 0 - Datum Index
            Index 1 - GroundTruth Index
            Index 2 - Prediction Index
            Index 3 - GroundTruth Label Index
            Index 4 - Prediction Label Index
            Index 5 - IOU
            Index 6 - Score
    label_metadata : NDArray[np.int32]
        Array containing label counts with shape (n_labels, 2)
            Index 0 - Ground truth label count
            Index 1 - Prediction label count

    Returns
    -------
    NDArray[np.float64]
        Array of ranked pairs for precision-recall metric computation.
    """
    pairs = detailed_pairs

    # remove null predictions
    pairs = pairs[pairs[:, 2] >= 0.0]

    # find best fits for prediction
    mask_label_match = np.isclose(pairs[:, 3], pairs[:, 4])
    matched_predictions = np.unique(pairs[mask_label_match, 2])
    mask_unmatched_predictions = ~np.isin(pairs[:, 2], matched_predictions)
    pairs = pairs[mask_label_match | mask_unmatched_predictions]

    # remove predictions for labels that have no ground truths
    for label_idx, count in enumerate(label_metadata[:, 0]):
        if count > 0:
            continue
        pairs = pairs[pairs[:, 4] != label_idx]

    # only keep the highest ranked pair
    _, indices = np.unique(pairs[:, [0, 2, 4]], axis=0, return_index=True)
    pairs = pairs[indices]

    # np.unique orders its results by value, we need to sort the indices to maintain the results of the lexsort
    indices = np.lexsort(
        (
            -pairs[:, 5],  # iou
            -pairs[:, 6],  # score
        )
    )
    pairs = pairs[indices]

    return pairs


def compute_precion_recall(
    ranked_pairs: NDArray[np.float64],
    label_metadata: NDArray[np.int32],
    iou_thresholds: NDArray[np.float64],
    score_thresholds: NDArray[np.float64],
) -> tuple[
    tuple[
        NDArray[np.float64],
        NDArray[np.float64],
    ],
    tuple[
        NDArray[np.float64],
        NDArray[np.float64],
    ],
    NDArray[np.float64],
    NDArray[np.float64],
]:
    """
    Computes Object Detection metrics.

    Takes data with shape (N, 7):

    Index 0 - Datum Index
    Index 1 - GroundTruth Index
    Index 2 - Prediction Index
    Index 3 - IOU
    Index 4 - GroundTruth Label Index
    Index 5 - Prediction Label Index
    Index 6 - Score

    Parameters
    ----------
    ranked_pairs : NDArray[np.float64]
        A ranked array summarizing the IOU calculations of one or more pairs.
    label_metadata : NDArray[np.int32]
        An array containing metadata related to labels.
    iou_thresholds : NDArray[np.float64]
        A 1-D array containing IOU thresholds.
    score_thresholds : NDArray[np.float64]
        A 1-D array containing score thresholds.

    Returns
    -------
    tuple[NDArray[np.float64], NDArray[np.float64]]
        Average Precision results (AP, mAP).
    tuple[NDArray[np.float64], NDArray[np.float64]]
        Average Recall results (AR, mAR).
    NDArray[np.float64]
        Precision, Recall, TP, FP, FN, F1 Score.
    NDArray[np.float64]
        Interpolated Precision-Recall Curves.
    """
    n_rows = ranked_pairs.shape[0]
    n_labels = label_metadata.shape[0]
    n_ious = iou_thresholds.shape[0]
    n_scores = score_thresholds.shape[0]

    # initialize result arrays
    average_precision = np.zeros((n_ious, n_labels), dtype=np.float64)
    mAP = np.zeros(n_ious, dtype=np.float64)
    average_recall = np.zeros((n_scores, n_labels), dtype=np.float64)
    mAR = np.zeros(n_scores, dtype=np.float64)
    counts = np.zeros((n_ious, n_scores, n_labels, 6), dtype=np.float64)
    pr_curve = np.zeros((n_ious, n_labels, 101, 2))

    # start computation
    ids = ranked_pairs[:, :5].astype(np.int32)
    gt_ids = ids[:, 1]
    gt_labels = ids[:, 3]
    pd_labels = ids[:, 4]
    ious = ranked_pairs[:, 5]
    scores = ranked_pairs[:, 6]

    unique_pd_labels, unique_pd_indices = np.unique(
        pd_labels, return_index=True
    )
    gt_count = label_metadata[:, 0]
    running_total_count = np.zeros(
        (n_ious, n_rows),
        dtype=np.float64,
    )
    running_tp_count = np.zeros_like(running_total_count)
    running_gt_count = np.zeros_like(running_total_count)

    mask_score_nonzero = scores > 1e-9
    mask_gt_exists = gt_ids >= 0.0
    mask_labels_match = np.isclose(gt_labels, pd_labels)

    mask_gt_exists_labels_match = mask_gt_exists & mask_labels_match

    mask_tp = mask_score_nonzero & mask_gt_exists_labels_match
    mask_fp = mask_score_nonzero
    mask_fn = mask_gt_exists_labels_match

    for iou_idx in range(n_ious):
        mask_iou = ious >= iou_thresholds[iou_idx]

        mask_tp_outer = mask_tp & mask_iou
        mask_fp_outer = mask_fp & (
            (~mask_gt_exists_labels_match & mask_iou) | ~mask_iou
        )
        mask_fn_outer = mask_fn & mask_iou

        for score_idx in range(n_scores):
            mask_score_thresh = scores >= score_thresholds[score_idx]

            mask_tp_inner = mask_tp_outer & mask_score_thresh
            mask_fp_inner = mask_fp_outer & mask_score_thresh
            mask_fn_inner = mask_fn_outer & ~mask_score_thresh

            # create true-positive mask score threshold
            tp_candidates = ids[mask_tp_inner]
            _, indices_gt_unique = np.unique(
                tp_candidates[:, [0, 1, 3]], axis=0, return_index=True
            )
            mask_gt_unique = np.zeros(tp_candidates.shape[0], dtype=np.bool_)
            mask_gt_unique[indices_gt_unique] = True

            true_positives_mask = np.zeros(n_rows, dtype=np.bool_)
            true_positives_mask[mask_tp_inner] = mask_gt_unique

            mask_fp_inner |= mask_tp_inner & ~true_positives_mask

            # calculate intermediates
            tp_count = np.bincount(
                pd_labels,
                weights=true_positives_mask,
                minlength=n_labels,
            ).astype(np.float64)
            fp_count = np.bincount(
                pd_labels[mask_fp_inner],
                minlength=n_labels,
            ).astype(np.float64)
            fn_count = np.bincount(
                pd_labels[mask_fn_inner],
                minlength=n_labels,
            )

            fn_count = gt_count - tp_count
            tp_fp_count = tp_count + fp_count

            # calculate component metrics
            recall = np.zeros_like(tp_count)
            np.divide(tp_count, gt_count, where=gt_count > 1e-9, out=recall)

            precision = np.zeros_like(tp_count)
            np.divide(
                tp_count, tp_fp_count, where=tp_fp_count > 1e-9, out=precision
            )

            f1_score = np.zeros_like(precision)
            np.divide(
                2 * np.multiply(precision, recall),
                (precision + recall),
                where=(precision + recall) > 1e-9,
                out=f1_score,
                dtype=np.float64,
            )

            counts[iou_idx][score_idx] = np.concatenate(
                (
                    tp_count[:, np.newaxis],
                    fp_count[:, np.newaxis],
                    fn_count[:, np.newaxis],
                    precision[:, np.newaxis],
                    recall[:, np.newaxis],
                    f1_score[:, np.newaxis],
                ),
                axis=1,
            )

            # calculate recall for AR
            average_recall[score_idx] += recall

        # create true-positive mask score threshold
        tp_candidates = ids[mask_tp_outer]
        _, indices_gt_unique = np.unique(
            tp_candidates[:, [0, 1, 3]], axis=0, return_index=True
        )
        mask_gt_unique = np.zeros(tp_candidates.shape[0], dtype=np.bool_)
        mask_gt_unique[indices_gt_unique] = True
        true_positives_mask = np.zeros(n_rows, dtype=np.bool_)
        true_positives_mask[mask_tp_outer] = mask_gt_unique

        # count running tp and total for AP
        for pd_label in unique_pd_labels:
            mask_pd_label = pd_labels == pd_label
            running_gt_count[iou_idx][mask_pd_label] = gt_count[pd_label]
            running_total_count[iou_idx][mask_pd_label] = np.arange(
                1, mask_pd_label.sum() + 1
            )
            mask_tp_for_counting = mask_pd_label & true_positives_mask
            running_tp_count[iou_idx][mask_tp_for_counting] = np.arange(
                1, mask_tp_for_counting.sum() + 1
            )

    # calculate running precision-recall points for AP
    precision = np.zeros_like(running_total_count)
    np.divide(
        running_tp_count,
        running_total_count,
        where=running_total_count > 1e-9,
        out=precision,
    )
    recall = np.zeros_like(running_total_count)
    np.divide(
        running_tp_count,
        running_gt_count,
        where=running_gt_count > 1e-9,
        out=recall,
    )
    recall_index = np.floor(recall * 100.0).astype(np.int32)

    # bin precision-recall curve
    for iou_idx in range(n_ious):
        p = precision[iou_idx]
        r = recall_index[iou_idx]
        pr_curve[iou_idx, pd_labels, r, 0] = np.maximum(
            pr_curve[iou_idx, pd_labels, r, 0],
            p,
        )
        pr_curve[iou_idx, pd_labels, r, 1] = np.maximum(
            pr_curve[iou_idx, pd_labels, r, 1],
            scores,
        )

    # calculate average precision
    running_max_precision = np.zeros((n_ious, n_labels), dtype=np.float64)
    running_max_score = np.zeros((n_labels), dtype=np.float64)
    for recall in range(100, -1, -1):

        # running max precision
        running_max_precision = np.maximum(
            pr_curve[:, :, recall, 0],
            running_max_precision,
        )
        pr_curve[:, :, recall, 0] = running_max_precision

        # running max score
        running_max_score = np.maximum(
            pr_curve[:, :, recall, 1],
            running_max_score,
        )
        pr_curve[:, :, recall, 1] = running_max_score

        average_precision += running_max_precision

    average_precision = average_precision / 101.0

    # calculate average recall
    average_recall = average_recall / n_ious

    # calculate mAP and mAR
    if unique_pd_labels.size > 0:
        mAP: NDArray[np.float64] = average_precision[:, unique_pd_labels].mean(
            axis=1
        )
        mAR: NDArray[np.float64] = average_recall[:, unique_pd_labels].mean(
            axis=1
        )

    return (
        (average_precision.astype(np.float64), mAP),
        (average_recall, mAR),
        counts,
        pr_curve,
    )


def _isin(
    data: NDArray,
    subset: NDArray,
) -> NDArray[np.bool_]:
    """
    Creates a mask of rows that exist within the subset.

    Parameters
    ----------
    data : NDArray[np.int32]
        An array with shape (N, 2).
    subset : NDArray[np.int32]
        An array with shape (M, 2) where N >= M.

    Returns
    -------
    NDArray[np.bool_]
        Returns a bool mask with shape (N,).
    """
    combined_data = (data[:, 0].astype(np.int64) << 32) | data[:, 1].astype(
        np.int32
    )
    combined_subset = (subset[:, 0].astype(np.int64) << 32) | subset[
        :, 1
    ].astype(np.int32)
    mask = np.isin(combined_data, combined_subset, assume_unique=False)
    return mask


class PairClassification(IntFlag):
    TP = auto()
    FP_FN_MISCLF = auto()
    FP_UNMATCHED = auto()
    FN_UNMATCHED = auto()


def mask_pairs_greedily(
    pairs: NDArray[np.float64],
):
    groundtruths = pairs[:, 1].astype(np.int32)
    predictions = pairs[:, 2].astype(np.int32)

    # Pre‑allocate "seen" flags for every possible x and y
    max_gt = groundtruths.max()
    max_pd = predictions.max()
    used_gt = np.zeros(max_gt + 1, dtype=np.bool_)
    used_pd = np.zeros(max_pd + 1, dtype=np.bool_)

    # This mask will mark which pairs to keep
    keep = np.zeros(pairs.shape[0], dtype=bool)

    for idx in range(groundtruths.shape[0]):
        gidx = groundtruths[idx]
        pidx = predictions[idx]

        if not (gidx < 0 or pidx < 0 or used_gt[gidx] or used_pd[pidx]):
            keep[idx] = True
            used_gt[gidx] = True
            used_pd[pidx] = True

    mask_matches = _isin(
        data=pairs[:, (1, 2)],
        subset=np.unique(pairs[np.ix_(keep, (1, 2))], axis=0),  # type: ignore - np.ix_ typing
    )

    return mask_matches


def compute_confusion_matrix(
    detailed_pairs: NDArray[np.float64],
    iou_thresholds: NDArray[np.float64],
    score_thresholds: NDArray[np.float64],
) -> NDArray[np.uint8]:
    """
    Compute detailed counts.

    Takes data with shape (N, 7):

    Index 0 - Datum Index
    Index 1 - GroundTruth Index
    Index 2 - Prediction Index
    Index 3 - GroundTruth Label Index
    Index 4 - Prediction Label Index
    Index 5 - IOU
    Index 6 - Score

    Parameters
    ----------
    detailed_pairs : NDArray[np.float64]
        An unsorted array summarizing the IOU calculations of one or more pairs.
    label_metadata : NDArray[np.int32]
        An array containing metadata related to labels.
    iou_thresholds : NDArray[np.float64]
        A 1-D array containing IOU thresholds.
    score_thresholds : NDArray[np.float64]
        A 1-D array containing score thresholds.

    Returns
    -------
    NDArray[np.uint8]
        Confusion matrix.
    """
    n_pairs = detailed_pairs.shape[0]
    n_ious = iou_thresholds.shape[0]
    n_scores = score_thresholds.shape[0]

    pair_classifications = np.zeros(
        (n_ious, n_scores, n_pairs),
        dtype=np.uint8,
    )

    ids = detailed_pairs[:, :5].astype(np.int32)
    groundtruths = ids[:, (0, 1)]
    predictions = ids[:, (0, 2)]
    gt_ids = ids[:, 1]
    pd_ids = ids[:, 2]
    gt_labels = ids[:, 3]
    pd_labels = ids[:, 4]
    ious = detailed_pairs[:, 5]
    scores = detailed_pairs[:, 6]

    mask_gt_exists = gt_ids > -0.5
    mask_pd_exists = pd_ids > -0.5
    mask_label_match = np.isclose(gt_labels, pd_labels)
    mask_score_nonzero = scores > 1e-9
    mask_iou_nonzero = ious > 1e-9

    mask_gt_pd_exists = mask_gt_exists & mask_pd_exists
    mask_gt_pd_match = mask_gt_pd_exists & mask_label_match

    mask_matched_pairs = mask_pairs_greedily(pairs=detailed_pairs)

    for iou_idx in range(n_ious):
        mask_iou_threshold = ious >= iou_thresholds[iou_idx]
        mask_iou = mask_iou_nonzero & mask_iou_threshold
        for score_idx in range(n_scores):
            mask_score_threshold = scores >= score_thresholds[score_idx]
            mask_score = mask_score_nonzero & mask_score_threshold

            mask_thresholded_matched_pairs = (
                mask_matched_pairs & mask_iou & mask_score
            )

            mask_true_positives = (
                mask_thresholded_matched_pairs & mask_gt_pd_match
            )
            mask_misclf = mask_thresholded_matched_pairs & ~mask_gt_pd_match

            mask_groundtruths_in_thresholded_matched_pairs = _isin(
                data=groundtruths,
                subset=np.unique(
                    groundtruths[mask_thresholded_matched_pairs], axis=0
                ),
            )
            mask_predictions_in_thresholded_matched_pairs = _isin(
                data=predictions,
                subset=np.unique(
                    predictions[mask_thresholded_matched_pairs], axis=0
                ),
            )

            mask_unmatched_predictions = (
                ~mask_predictions_in_thresholded_matched_pairs
                & mask_pd_exists
                & mask_score
            )
            mask_unmatched_groundtruths = (
                ~mask_groundtruths_in_thresholded_matched_pairs
                & mask_gt_exists
            )

            # classify pairings
            pair_classifications[
                iou_idx, score_idx, mask_true_positives
            ] |= np.uint8(PairClassification.TP)
            pair_classifications[iou_idx, score_idx, mask_misclf] |= np.uint8(
                PairClassification.FP_FN_MISCLF
            )
            pair_classifications[
                iou_idx, score_idx, mask_unmatched_predictions
            ] |= np.uint8(PairClassification.FP_UNMATCHED)
            pair_classifications[
                iou_idx, score_idx, mask_unmatched_groundtruths
            ] |= np.uint8(PairClassification.FN_UNMATCHED)

    return pair_classifications



from dataclasses import asdict, dataclass

import numpy as np
from numpy.typing import NDArray
from tqdm import tqdm

from valor_lite.classification.annotation import Classification
from valor_lite.classification.computation import (
    compute_confusion_matrix,
    compute_label_metadata,
    compute_precision_recall_rocauc,
    filter_cache,
)
from valor_lite.classification.metric import Metric, MetricType
from valor_lite.classification.utilities import (
    unpack_confusion_matrix_into_metric_list,
    unpack_precision_recall_rocauc_into_metric_lists,
)
from valor_lite.exceptions import EmptyEvaluatorError, EmptyFilterError

"""
Usage
-----

manager = DataLoader()
manager.add_data(
    groundtruths=groundtruths,
    predictions=predictions,
)
evaluator = manager.finalize()

metrics = evaluator.evaluate()

f1_metrics = metrics[MetricType.F1]
accuracy_metrics = metrics[MetricType.Accuracy]

filter_mask = evaluator.create_filter(datum_uids=["uid1", "uid2"])
filtered_metrics = evaluator.evaluate(filter_mask=filter_mask)
"""


@dataclass
class Metadata:
    number_of_datums: int = 0
    number_of_ground_truths: int = 0
    number_of_predictions: int = 0
    number_of_labels: int = 0

    @classmethod
    def create(
        cls,
        detailed_pairs: NDArray[np.float64],
        number_of_datums: int,
        number_of_labels: int,
    ):
        # count number of unique ground truths
        mask_valid_gts = detailed_pairs[:, 1] >= 0
        unique_ids = np.unique(
            detailed_pairs[np.ix_(mask_valid_gts, (0, 1))],  # type: ignore - np.ix_ typing
            axis=0,
        )
        number_of_ground_truths = int(unique_ids.shape[0])

        # count number of unqiue predictions
        mask_valid_pds = detailed_pairs[:, 2] >= 0
        unique_ids = np.unique(
            detailed_pairs[np.ix_(mask_valid_pds, (0, 2))], axis=0  # type: ignore - np.ix_ typing
        )
        number_of_predictions = int(unique_ids.shape[0])

        return cls(
            number_of_datums=number_of_datums,
            number_of_ground_truths=number_of_ground_truths,
            number_of_predictions=number_of_predictions,
            number_of_labels=number_of_labels,
        )

    def to_dict(self) -> dict[str, int | bool]:
        return asdict(self)


@dataclass
class Filter:
    datum_mask: NDArray[np.bool_]
    valid_label_indices: NDArray[np.int32] | None
    metadata: Metadata

    def __post_init__(self):
        # validate datum mask
        if not self.datum_mask.any():
            raise EmptyFilterError("filter removes all datums")

        # validate label indices
        if (
            self.valid_label_indices is not None
            and self.valid_label_indices.size == 0
        ):
            raise EmptyFilterError("filter removes all labels")


class Evaluator:
    """
    Classification Evaluator
    """

    def __init__(self):
        # external references
        self.datum_id_to_index: dict[str, int] = {}
        self.label_to_index: dict[str, int] = {}

        self.index_to_datum_id: list[str] = []
        self.index_to_label: list[str] = []

        # internal caches
        self._detailed_pairs = np.array([])
        self._label_metadata = np.array([], dtype=np.int32)
        self._metadata = Metadata()

    @property
    def metadata(self) -> Metadata:
        return self._metadata

    @property
    def ignored_prediction_labels(self) -> list[str]:
        """
        Prediction labels that are not present in the ground truth set.
        """
        glabels = set(np.where(self._label_metadata[:, 0] > 0)[0])
        plabels = set(np.where(self._label_metadata[:, 1] > 0)[0])
        return [
            self.index_to_label[label_id] for label_id in (plabels - glabels)
        ]

    @property
    def missing_prediction_labels(self) -> list[str]:
        """
        Ground truth labels that are not present in the prediction set.
        """
        glabels = set(np.where(self._label_metadata[:, 0] > 0)[0])
        plabels = set(np.where(self._label_metadata[:, 1] > 0)[0])
        return [
            self.index_to_label[label_id] for label_id in (glabels - plabels)
        ]

    def create_filter(
        self,
        datums: list[str] | NDArray[np.int32] | None = None,
        labels: list[str] | NDArray[np.int32] | None = None,
    ) -> Filter:
        """
        Creates a filter object.

        Parameters
        ----------
        datums : list[str] | NDArray[int32], optional
            An optional list of string uids or integer indices representing datums.
        labels : list[str] | NDArray[int32], optional
            An optional list of strings or integer indices representing labels.

        Returns
        -------
        Filter
            The filter object representing the input parameters.
        """
        # create datum mask
        n_pairs = self._detailed_pairs.shape[0]
        datum_mask = np.ones(n_pairs, dtype=np.bool_)
        if datums is not None:
            # convert to array of valid datum indices
            if isinstance(datums, list):
                datums = np.array(
                    [self.datum_id_to_index[uid] for uid in datums],
                    dtype=np.int32,
                )

            # return early if all data removed
            if datums.size == 0:
                raise EmptyFilterError("filter removes all datums")

            # validate indices
            if datums.max() >= len(self.index_to_datum_id):
                raise ValueError(
                    f"datum index '{datums.max()}' exceeds total number of datums"
                )
            elif datums.min() < 0:
                raise ValueError(
                    f"datum index '{datums.min()}' is a negative value"
                )

            # create datum mask
            datum_mask = np.isin(self._detailed_pairs[:, 0], datums)

        # collect valid label indices
        if labels is not None:
            # convert to array of valid label indices
            if isinstance(labels, list):
                labels = np.array(
                    [self.label_to_index[label] for label in labels]
                )

            # return early if all data removed
            if labels.size == 0:
                raise EmptyFilterError("filter removes all labels")

            # validate indices
            if labels.max() >= len(self.index_to_label):
                raise ValueError(
                    f"label index '{labels.max()}' exceeds total number of labels"
                )
            elif labels.min() < 0:
                raise ValueError(
                    f"label index '{labels.min()}' is a negative value"
                )

            # add -1 to represent null labels which should not be filtered
            labels = np.concatenate([labels, np.array([-1])])

        filtered_detailed_pairs, _ = filter_cache(
            detailed_pairs=self._detailed_pairs,
            datum_mask=datum_mask,
            valid_label_indices=labels,
            n_labels=self.metadata.number_of_labels,
        )

        number_of_datums = (
            datums.size
            if datums is not None
            else self.metadata.number_of_datums
        )

        return Filter(
            datum_mask=datum_mask,
            valid_label_indices=labels,
            metadata=Metadata.create(
                detailed_pairs=filtered_detailed_pairs,
                number_of_datums=number_of_datums,
                number_of_labels=self.metadata.number_of_labels,
            ),
        )

    def filter(
        self, filter_: Filter
    ) -> tuple[NDArray[np.float64], NDArray[np.int32]]:
        """
        Performs filtering over the internal cache.

        Parameters
        ----------
        filter_ : Filter
            The filter object representation.

        Returns
        -------
        NDArray[float64]
            The filtered detailed pairs.
        NDArray[int32]
            The filtered label metadata.
        """
        return filter_cache(
            detailed_pairs=self._detailed_pairs,
            datum_mask=filter_.datum_mask,
            valid_label_indices=filter_.valid_label_indices,
            n_labels=self.metadata.number_of_labels,
        )

    def compute_precision_recall_rocauc(
        self,
        score_thresholds: list[float] = [0.0],
        hardmax: bool = True,
        filter_: Filter | None = None,
    ) -> dict[MetricType, list]:
        """
        Performs an evaluation and returns metrics.

        Parameters
        ----------
        score_thresholds : list[float]
            A list of score thresholds to compute metrics over.
        hardmax : bool
            Toggles whether a hardmax is applied to predictions.
        filter_ : Filter, optional
            Applies a filter to the internal cache.

        Returns
        -------
        dict[MetricType, list]
            A dictionary mapping MetricType enumerations to lists of computed metrics.
        """
        # apply filters
        if filter_ is not None:
            detailed_pairs, label_metadata = self.filter(filter_=filter_)
            n_datums = filter_.metadata.number_of_datums
        else:
            detailed_pairs = self._detailed_pairs
            label_metadata = self._label_metadata
            n_datums = self.metadata.number_of_datums

        results = compute_precision_recall_rocauc(
            detailed_pairs=detailed_pairs,
            label_metadata=label_metadata,
            score_thresholds=np.array(score_thresholds),
            hardmax=hardmax,
            n_datums=n_datums,
        )
        return unpack_precision_recall_rocauc_into_metric_lists(
            results=results,
            score_thresholds=score_thresholds,
            hardmax=hardmax,
            label_metadata=label_metadata,
            index_to_label=self.index_to_label,
        )

    def compute_confusion_matrix(
        self,
        score_thresholds: list[float] = [0.0],
        hardmax: bool = True,
        filter_: Filter | None = None,
    ) -> list[Metric]:
        """
        Computes a detailed confusion matrix..

        Parameters
        ----------
        score_thresholds : list[float]
            A list of score thresholds to compute metrics over.
        hardmax : bool
            Toggles whether a hardmax is applied to predictions.
        filter_ : Filter, optional
            Applies a filter to the internal cache.

        Returns
        -------
        list[Metric]
            A list of confusion matrices.
        """
        # apply filters
        if filter_ is not None:
            detailed_pairs, _ = self.filter(filter_=filter_)
        else:
            detailed_pairs = self._detailed_pairs

        if detailed_pairs.size == 0:
            return list()

        result = compute_confusion_matrix(
            detailed_pairs=detailed_pairs,
            score_thresholds=np.array(score_thresholds),
            hardmax=hardmax,
        )
        return unpack_confusion_matrix_into_metric_list(
            detailed_pairs=detailed_pairs,
            result=result,
            score_thresholds=score_thresholds,
            index_to_datum_id=self.index_to_datum_id,
            index_to_label=self.index_to_label,
        )

    def evaluate(
        self,
        score_thresholds: list[float] = [0.0],
        hardmax: bool = True,
        filter_: Filter | None = None,
    ) -> dict[MetricType, list[Metric]]:
        """
        Computes a detailed confusion matrix..

        Parameters
        ----------
        score_thresholds : list[float]
            A list of score thresholds to compute metrics over.
        hardmax : bool
            Toggles whether a hardmax is applied to predictions.
        filter_ : Filter, optional
            Applies a filter to the internal cache.

        Returns
        -------
        dict[MetricType, list[Metric]]
            Lists of metrics organized by metric type.
        """
        metrics = self.compute_precision_recall_rocauc(
            score_thresholds=score_thresholds,
            hardmax=hardmax,
            filter_=filter_,
        )
        metrics[MetricType.ConfusionMatrix] = self.compute_confusion_matrix(
            score_thresholds=score_thresholds,
            hardmax=hardmax,
            filter_=filter_,
        )
        return metrics

    def _add_datum(self, uid: str) -> int:
        """
        Helper function for adding a datum to the cache.

        Parameters
        ----------
        uid : str
            The datum uid.

        Returns
        -------
        int
            The datum index.

        Raises
        ------
        ValueError
            If datum id already exists.
        """
        if uid in self.datum_id_to_index:
            raise ValueError("datum with id '{uid}' already exists")
        index = len(self.datum_id_to_index)
        self.datum_id_to_index[uid] = index
        self.index_to_datum_id.append(uid)
        return self.datum_id_to_index[uid]

    def _add_label(self, label: str) -> int:
        """
        Helper function for adding a label to the cache.

        Parameters
        ----------
        label : str
            A string representing a label.

        Returns
        -------
        int
            Label index.
        """
        label_id = len(self.index_to_label)
        if label not in self.label_to_index:
            self.label_to_index[label] = label_id
            self.index_to_label.append(label)
            label_id += 1
        return self.label_to_index[label]

    def add_data(
        self,
        classifications: list[Classification],
        show_progress: bool = False,
    ):
        """
        Adds classifications to the cache.

        Parameters
        ----------
        classifications : list[Classification]
            A list of Classification objects.
        show_progress : bool, default=False
            Toggle for tqdm progress bar.
        """

        disable_tqdm = not show_progress
        for classification in tqdm(classifications, disable=disable_tqdm):

            if len(classification.predictions) == 0:
                raise ValueError(
                    "Classifications must contain at least one prediction."
                )

            # update datum uid index
            uid_index = self._add_datum(uid=classification.uid)

            # cache labels and annotations
            groundtruth = self._add_label(classification.groundtruth)

            predictions = list()
            for plabel, pscore in zip(
                classification.predictions, classification.scores
            ):
                label_idx = self._add_label(plabel)
                predictions.append(
                    (
                        label_idx,
                        pscore,
                    )
                )

            pairs = list()
            scores = np.array([score for _, score in predictions])
            max_score_idx = np.argmax(scores)

            for idx, (plabel, score) in enumerate(predictions):
                pairs.append(
                    (
                        float(uid_index),
                        float(groundtruth),
                        float(plabel),
                        float(score),
                        float(max_score_idx == idx),
                    )
                )

            if self._detailed_pairs.size == 0:
                self._detailed_pairs = np.array(pairs)
            else:
                self._detailed_pairs = np.concatenate(
                    [
                        self._detailed_pairs,
                        np.array(pairs),
                    ],
                    axis=0,
                )

    def finalize(self):
        """
        Performs data finalization and some preprocessing steps.

        Returns
        -------
        Evaluator
            A ready-to-use evaluator object.
        """
        if self._detailed_pairs.size == 0:
            raise EmptyEvaluatorError()

        self._label_metadata = compute_label_metadata(
            ids=self._detailed_pairs[:, :3].astype(np.int32),
            n_labels=len(self.index_to_label),
        )
        indices = np.lexsort(
            (
                self._detailed_pairs[:, 1],  # ground truth
                self._detailed_pairs[:, 2],  # prediction
                -self._detailed_pairs[:, 3],  # score
            )
        )
        self._detailed_pairs = self._detailed_pairs[indices]
        self._metadata = Metadata.create(
            detailed_pairs=self._detailed_pairs,
            number_of_datums=len(self.index_to_datum_id),
            number_of_labels=len(self.index_to_label),
        )
        return self


class DataLoader(Evaluator):
    """
    Used for backwards compatibility as the Evaluator now handles ingestion.
    """

    pass
